import inspect
from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, List

from hydrolib.core.basemodel import SerializerConfig


class BuiEventSerializer:
    """
    Serializer class to transform a bui event into a text block.
    """

    bui_event_template = inspect.cleandoc(
        """
        * Event {event_idx} duration days:{d_days} hours:{d_hours} minutes:{d_minutes} seconds:{d_seconds}
        * Start date and time of the event: yyyy mm dd hh mm ss
        * Duration of the event           : dd hh mm ss
        * Rainfall value per time step [mm/time step]
        {start_time} {timeseries_length}
        {precipitation_per_timestep}
    """
    )

    @staticmethod
    def serialize(event_data: Dict, config: SerializerConfig) -> str:
        """
        Serializes a dictionary representing an event into a text block.

        Args:
            event_data (Dict): Dictionary representing precipitation event.
            config (SerializerConfig): The serialization configuration.

        Returns:
            str: Formatted string.
        """
        event_data["start_time"] = BuiEventSerializer.serialize_start_time(
            event_data["start_time"]
        )
        ts_duration = event_data["timeseries_length"]
        event_data = {
            **event_data,
            **BuiEventSerializer.get_timedelta_fields(ts_duration),
        }
        event_data[
            "timeseries_length"
        ] = BuiEventSerializer.serialize_timeseries_length(
            event_data["timeseries_length"]
        )
        event_data[
            "precipitation_per_timestep"
        ] = BuiEventSerializer.serialize_precipitation_per_timestep(
            event_data["precipitation_per_timestep"], config
        )
        if "event_idx" not in event_data.keys():
            event_data["event_idx"] = 1
        return BuiEventSerializer.bui_event_template.format(**event_data)

    @staticmethod
    def get_timedelta_fields(duration: timedelta) -> Dict:
        """
        Gets a dictionary containing the time delta in days, hours, minutes and seconds.
        This means that the seconds field does not contain the accumulative value of days
        hours and minutes.

        Args:
            duration (timedelta): Timedelta to convert.

        Returns:
            Dict: Dictionary containing all fields.
        """
        total_hours = int(duration.seconds / (60 * 60))
        total_minutes = int((duration.seconds / 60) - (total_hours * 60))
        total_seconds = int(
            duration.seconds - ((total_hours * 60 + total_minutes) * 60)
        )
        return dict(
            d_seconds=total_seconds,
            d_minutes=total_minutes,
            d_hours=total_hours,
            d_days=duration.days,
        )

    @staticmethod
    def serialize_start_time(data_to_serialize: datetime) -> str:
        """
        Serializes a datetime into the expected .bui format.

        Args:
            data_to_serialize (datetime): Datetime representing reference time.

        Returns:
            str: Converted datetime into string.
        """
        # Not using the following format because we only want one digit instead of
        # double (day 1 -> 1, instead of 01).
        # data_to_serialize.strftime("%Y %m %d %H %M %S")
        dt = data_to_serialize
        return f"{dt.year} {dt.month} {dt.day} {dt.hour} {dt.minute} {dt.second}"

    @staticmethod
    def serialize_timeseries_length(data_to_serialize: timedelta) -> str:
        """
        Serializes a given timedelta into the .bui format.

        Args:
            data_to_serialize (timedelta): Reference timespan to serialize.

        Returns:
            str: Converted timedelta in string.
        """
        fields_dict = BuiEventSerializer.get_timedelta_fields(data_to_serialize)
        total_hours = fields_dict["d_hours"]
        total_minutes = fields_dict["d_minutes"]
        total_seconds = fields_dict["d_seconds"]
        return f"{data_to_serialize.days} {total_hours} {total_minutes} {total_seconds}"

    @staticmethod
    def serialize_precipitation_per_timestep(
        data_to_serialize: List[List[float]], config: SerializerConfig
    ) -> str:
        """
        Serialized the data containing all the precipitations per timestep (and station)
        into a single string ready to be mapped.

        Args:
            data_to_serialize (List[List[str]]): Data to be mapped.
            config (SerializerConfig): The serialization configuration.

        Returns:
            str: Serialized string in .bui format.
        """
        float_format = lambda v: f"{v:{config.float_format}}"
        serialized_data = str.join(
            "\n",
            [
                str.join(" ", map(float_format, listed_data))
                for listed_data in data_to_serialize
            ],
        )
        return serialized_data


class BuiSerializer:
    """
    Serializer class to transform an object into a .bui file text format.
    """

    bui_template = inspect.cleandoc(
        """
        *Name of this file: {filepath}
        *Date and time of construction: {datetime_now}
        *Comments are following an * (asterisk) and written above variables
        {default_dataset}
        *Number of stations
        {number_of_stations}
        *Station Name
        {name_of_stations}
        *Number_of_events seconds_per_timestamp
        {number_of_events} {seconds_per_timestep}
        {precipitation_events}
        """
    )

    @staticmethod
    def serialize(bui_data: Dict, config: SerializerConfig) -> str:
        """
        Formats the bui_template with the content of the given data.
        NOTE: It requires that caller injects file_path into bui_data prior to this call.
        Otherwise it will crash.

        Args:
            bui_data (Dict): Data to serialize.
            config (SerializerConfig): The serialization configuration.

        Returns:
            str: The serialized data.
        """
        bui_data["datetime_now"] = datetime.now().strftime("%d-%m-%y %H:%M:%S")
        bui_data["name_of_stations"] = BuiSerializer.serialize_stations_ids(
            bui_data["name_of_stations"]
        )
        bui_data["precipitation_events"] = BuiSerializer.serialize_event_list(
            bui_data["precipitation_events"], config
        )
        return BuiSerializer.bui_template.format(**bui_data)

    @staticmethod
    def serialize_event_list(
        data_to_serialize: List[Dict], config: SerializerConfig
    ) -> str:
        """
        Serializes a event list dictionary into a single text block.

        Args:
            data_to_serialize (Dict): Dictionary containing list of events.
            config (SerializerConfig): The serialization configuration.

        Returns:
            str: Text block representing all precipitation events.
        """
        serialized_list = []
        for n_event, event in enumerate(data_to_serialize):
            event["event_idx"] = n_event + 1
            serialized_list.append(BuiEventSerializer.serialize(event, config))
        return "\n".join(serialized_list)

    @staticmethod
    def serialize_stations_ids(data_to_serialize: List[str]) -> str:
        """
        Serializes the stations ids into a single string as expected in a .bui file.

        Args:
            data_to_serialize (List[str]): List of station ids.

        Returns:
            str: Serialized string.
        """
        return str.join(" ", data_to_serialize)


def write_bui_file(path: Path, data: Dict, config: SerializerConfig) -> None:
    """
    Writes a .bui file in the given path based on the data given in a dictionary.

    Args:
        path (Path): Path where to output the text.
        data (Dict): Data to serialize into the file.
        config (SerializerConfig): The serialization configuration.
    """
    data["filepath"] = path  # This is redundant as already exists in the data.
    serialized_bui_data = BuiSerializer.serialize(data, config)

    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(serialized_bui_data, encoding="utf8")



"""
structure namespace for storing the contents of an [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]'s structure file.
"""
# TODO Implement the following structures
# - Gate

import logging
from enum import Enum
from operator import gt, ne
from pathlib import Path
from typing import Dict, List, Literal, Optional, Set, Union

from pydantic import Field
from pydantic.class_validators import root_validator, validator

from hydrolib.core.basemodel import DiskOnlyFileModel
from hydrolib.core.dflowfm.friction.models import FrictionType
from hydrolib.core.dflowfm.ini.models import INIBasedModel, INIGeneral, INIModel
from hydrolib.core.dflowfm.ini.util import (
    get_enum_validator,
    get_from_subclass_defaults,
    get_split_string_on_delimiter_validator,
    make_list_validator,
    validate_conditionally,
    validate_correct_length,
    validate_forbidden_fields,
    validate_required_fields,
)
from hydrolib.core.utils import str_is_empty_or_none

logger = logging.getLogger(__name__)

# TODO: handle comment blocks
# TODO: handle duplicate keys
class Structure(INIBasedModel):
    # TODO: would we want to load this from something externally and generate these automatically
    class Comments(INIBasedModel.Comments):
        id: Optional[str] = "Unique structure id (max. 256 characters)."
        name: Optional[str] = "Given name in the user interface."
        polylinefile: Optional[str] = Field(
            "*.pli; Polyline geometry definition for 2D structure.",
            alias="polylinefile",
        )
        branchid: Optional[str] = Field(
            "Branch on which the structure is located.", alias="branchId"
        )
        chainage: Optional[str] = "Chainage on the branch (m)."

        numcoordinates: Optional[str] = Field(
            "Number of values in xCoordinates and yCoordinates", alias="numCoordinates"
        )
        xcoordinates: Optional[str] = Field(
            "x-coordinates of the location of the structure. (number of values = numCoordinates)",
            alias="xCoordinates",
        )
        ycoordinates: Optional[str] = Field(
            "y-coordinates of the location of the structure. (number of values = numCoordinates)",
            alias="yCoordinates",
        )

    comments: Comments = Comments()

    _header: Literal["Structure"] = "Structure"

    id: str = Field("id", max_length=256, alias="id")
    name: str = Field("id", alias="name")
    type: str = Field(alias="type")

    polylinefile: Optional[DiskOnlyFileModel] = Field(None, alias="polylinefile")

    branchid: Optional[str] = Field(None, alias="branchId")
    chainage: Optional[float] = Field(None, alias="chainage")

    numcoordinates: Optional[int] = Field(None, alias="numCoordinates")
    xcoordinates: Optional[List[float]] = Field(None, alias="xCoordinates")
    ycoordinates: Optional[List[float]] = Field(None, alias="yCoordinates")

    _loc_coord_fields = {"numcoordinates", "xcoordinates", "ycoordinates"}
    _loc_branch_fields = {"branchid", "chainage"}
    _loc_all_fields = _loc_coord_fields | _loc_branch_fields

    _split_to_list = get_split_string_on_delimiter_validator(
        "xcoordinates", "ycoordinates"
    )

    @validator("type", pre=True)
    def _validate_type(cls, value):
        return get_from_subclass_defaults(Structure, "type", value)

    @root_validator
    @classmethod
    def check_location(cls, values: dict) -> dict:
        """
        Validates the location of the structure based on the given parameters.
        For instance, if a branchid is given, then it is expected also the chainage,
        otherwise numcoordinates xcoordinates and ycoordinates shall be expected.

        Args:
            values (dict): Dictionary of values validated for the new structure.

        Raises:
            ValueError: When branchid or chainage values are not valid (empty strings).
            ValueError: When the number of xcoordinates and ycoordinates do not match numcoordinates.

        Returns:
            dict: Dictionary of values validated for the new structure.
        """
        filtered_values = {k: v for k, v in values.items() if v is not None}
        structype = filtered_values.get("type", "").lower()

        if structype == "compound" or issubclass(cls, (Compound)):
            # Compound structure does not require a location specification.
            return values

        # Backwards compatibility for old-style polylinefile input field (instead of num/x/yCoordinates):
        polyline_compatible_structures = dict(
            pump="Pump",
            dambreak="Dambreak",
            gate="Gate",
            weir="Weir",
            generalstructure="GeneralStructure",
        )
        polylinefile_in_model = (
            structype in polyline_compatible_structures.keys()
            and filtered_values.get("polylinefile") is not None
        )

        # No branchId+chainage for some structures:
        only_coordinates_structures = dict(
            longculvert="LongCulvert", dambreak="Dambreak"
        )
        coordinates_in_model = Structure.validate_coordinates_in_model(filtered_values)

        # Error: do not allow both x/y and polyline file:
        assert not (
            polylinefile_in_model and coordinates_in_model
        ), f"`Specify location either by `num/x/yCoordinates` or `polylinefile`, but not both."

        # Error: require x/y or polyline file:
        if (
            structype in polyline_compatible_structures.keys()
            and structype in only_coordinates_structures.keys()
        ):
            assert (
                coordinates_in_model or polylinefile_in_model
            ), f"Specify location either by setting `num/x/yCoordinates` or `polylinefile` fields for a {polyline_compatible_structures[structype]} structure."

        # Error: Some structures require coordinates_in_model, but not branchId and chainage.
        if (
            not polylinefile_in_model
            and structype in only_coordinates_structures.keys()
        ):
            assert (
                coordinates_in_model
            ), f"Specify location by setting `num/x/yCoordinates` for a {only_coordinates_structures[structype]} structure."

        # Error: final check: at least one of x/y, branchId+chainage or polyline file must be given
        branch_and_chainage_in_model = Structure.validate_branch_and_chainage_in_model(
            filtered_values
        )
        assert (
            branch_and_chainage_in_model
            or coordinates_in_model
            or polylinefile_in_model
        ), "Specify location either by setting `branchId` and `chainage` or `num/x/yCoordinates` or `polylinefile` fields."

        return values

    @staticmethod
    def validate_branch_and_chainage_in_model(values: dict) -> bool:
        """
        Static method to validate whether the given branchid and chainage values
        match the expectation of a new structure.

        Args:
            values (dict): Dictionary of values to be used to generate a structure.

        Raises:
            ValueError: When the value for branchid or chainage are not valid.

        Returns:
            bool: Result of valid branchid / chainage in dictionary.
        """
        branchid = values.get("branchid", None)
        if branchid is None:
            return False

        chainage = values.get("chainage", None)
        if str_is_empty_or_none(branchid) or chainage is None:
            raise ValueError(
                "A valid value for branchId and chainage is required when branchId key is specified."
            )
        return True

    @staticmethod
    def validate_coordinates_in_model(values: dict) -> bool:
        """
        Static method to validate whether the given values match the expectations
        of a structure to define its coordinates.

        Args:
            values (dict): Dictionary of values to be used to generate a structure.

        Raises:
            ValueError: When the given coordinates is less than 2.
            ValueError: When the given coordinates do not match in expected size.

        Returns:
            bool: Result of valid coordinates in dictionary.
        """
        searched_keys = ["numcoordinates", "xcoordinates", "ycoordinates"]
        if any(values.get(k, None) is None for k in searched_keys):
            return False

        n_coords = values["numcoordinates"]
        if n_coords < 2:
            raise ValueError(
                f"Expected at least 2 coordinates, but only {n_coords} declared."
            )

        def get_coord_len(coord: str) -> int:
            if values[coord] is None:
                return 0
            return len(values[coord])

        len_x_coords = get_coord_len("xcoordinates")
        len_y_coords = get_coord_len("ycoordinates")
        if n_coords == len_x_coords == len_y_coords:
            return True
        raise ValueError(
            f"Expected {n_coords} coordinates, given {len_x_coords} for xCoordinates and {len_y_coords} for yCoordinates."
        )

    @classmethod
    def validate(cls, v):
        """Try to initialize subclass based on the `type` field.
        This field is compared to each `type` field of the derived models of `Structure`.
        The derived model with an equal structure type will be initialized.

        Raises:
            ValueError: When the given type is not a known structure type.
        """

        # should be replaced by discriminated unions once merged
        # https://github.com/samuelcolvin/pydantic/pull/2336
        if isinstance(v, dict):
            for c in cls.__subclasses__():
                if (
                    c.__fields__.get("type").default.lower()
                    == v.get("type", "").lower()
                ):
                    v = c(**v)
                    break
            else:
                raise ValueError(
                    f"Type of {cls.__name__} with id={v.get('id', '')} and type={v.get('type', '')} is not recognized."
                )
        return super().validate(v)

    def _exclude_fields(self) -> Set:
        # exclude the non-applicable, or unset props like coordinates or branches
        if self.type == "compound":
            exclude_set = self._loc_all_fields
        elif self.branchid is not None:
            exclude_set = self._loc_coord_fields
        else:
            exclude_set = self._loc_branch_fields
        exclude_set = super()._exclude_fields().union(exclude_set)
        return exclude_set

    def _get_identifier(self, data: dict) -> Optional[str]:
        return data.get("id") or data.get("name")


class FlowDirection(str, Enum):
    """
    Enum class containing the valid values for the allowedFlowDirection
    attribute in several subclasses of Structure.
    """

    none = "none"
    positive = "positive"
    negative = "negative"
    both = "both"
    allowedvaluestext = "Possible values: both, positive, negative, none."


class Orientation(str, Enum):
    """
    Enum class containing the valid values for the orientation
    attribute in several subclasses of Structure.
    """

    positive = "positive"
    negative = "negative"
    allowedvaluestext = "Possible values: positive, negative."


class Weir(Structure):
    """
    Hydraulic structure with `type=weir`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the weir input as described in
    [UM Sec.C.12.1](https://content.oss.deltares.nl/delft3d/manuals/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.1).
    """

    class Comments(Structure.Comments):
        type: Optional[str] = Field("Structure type; must read weir", alias="type")
        allowedflowdir: Optional[str] = Field(
            FlowDirection.allowedvaluestext, alias="allowedFlowdir"
        )

        crestlevel: Optional[str] = Field(
            "Crest level of weir (m AD).", alias="crestLevel"
        )
        crestwidth: Optional[str] = Field("Width of the weir (m).", alias="crestWidth")
        corrcoeff: Optional[str] = Field(
            "Correction coefficient (-).", alias="corrCoeff"
        )
        usevelocityheight: Optional[str] = Field(
            "Flag indicating whether the velocity height is to be calculated or not.",
            alias="useVelocityHeight",
        )

    comments: Comments = Comments()

    type: Literal["weir"] = Field("weir", alias="type")
    allowedflowdir: Optional[FlowDirection] = Field(
        FlowDirection.both.value, alias="allowedFlowDir"
    )

    crestlevel: Union[float, Path] = Field(alias="crestLevel")
    crestwidth: Optional[float] = Field(None, alias="crestWidth")
    corrcoeff: float = Field(1.0, alias="corrCoeff")
    usevelocityheight: bool = Field(True, alias="useVelocityHeight")

    _flowdirection_validator = get_enum_validator("allowedflowdir", enum=FlowDirection)


class UniversalWeir(Structure):
    """
    Hydraulic structure with `type=universalWeir`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the universal weir input as described in
    [UM Sec.C.12.2](https://content.oss.deltares.nl/delft3d/manuals/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.2).
    """

    class Comments(Structure.Comments):
        type: Optional[str] = Field(
            "Structure type; must read universalWeir", alias="type"
        )
        allowedflowdir: Optional[str] = Field(
            FlowDirection.allowedvaluestext, alias="allowedFlowdir"
        )

        numlevels: Optional[str] = Field("Number of yz-Values.", alias="numLevels")
        yvalues: Optional[str] = Field(
            "y-values of the cross section (m). (number of values = numLevels)",
            alias="yValues",
        )
        zvalues: Optional[str] = Field(
            "z-values of the cross section (m). (number of values = numLevels)",
            alias="zValues",
        )
        crestlevel: Optional[str] = Field(
            "Crest level of weir (m AD).", alias="crestLevel"
        )
        dischargecoeff: Optional[str] = Field(
            "Discharge coefficient c_e (-).", alias="dischargeCoeff"
        )

    comments: Comments = Comments()

    type: Literal["universalWeir"] = Field("universalWeir", alias="type")
    allowedflowdir: FlowDirection = Field(alias="allowedFlowDir")

    numlevels: int = Field(alias="numLevels")
    yvalues: List[float] = Field(alias="yValues")
    zvalues: List[float] = Field(alias="zValues")
    crestlevel: float = Field(alias="crestLevel")
    dischargecoeff: float = Field(alias="dischargeCoeff")

    _split_to_list = get_split_string_on_delimiter_validator("yvalues", "zvalues")
    _flowdirection_validator = get_enum_validator("allowedflowdir", enum=FlowDirection)


class CulvertSubType(str, Enum):
    """Enum class to store a [Culvert][hydrolib.core.dflowfm.structure.models.Culvert]'s subType."""

    culvert = "culvert"
    invertedSiphon = "invertedSiphon"


class Culvert(Structure):
    """
    Hydraulic structure with `type=culvert`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the culvert input as described in
    [UM Sec.C.12.3](https://content.oss.deltares.nl/delft3d/manuals/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.3).
    """

    type: Literal["culvert"] = Field("culvert", alias="type")
    allowedflowdir: FlowDirection = Field(alias="allowedFlowDir")

    leftlevel: float = Field(alias="leftLevel")
    rightlevel: float = Field(alias="rightLevel")
    csdefid: str = Field(alias="csDefId")
    length: float = Field(alias="length")
    inletlosscoeff: float = Field(alias="inletLossCoeff")
    outletlosscoeff: float = Field(alias="outletLossCoeff")
    valveonoff: bool = Field(alias="valveOnOff")
    valveopeningheight: Optional[Union[float, Path]] = Field(alias="valveOpeningHeight")
    numlosscoeff: Optional[int] = Field(alias="numLossCoeff")
    relopening: Optional[List[float]] = Field(alias="relOpening")
    losscoeff: Optional[List[float]] = Field(alias="lossCoeff")
    bedfrictiontype: Optional[FrictionType] = Field(alias="bedFrictionType")
    bedfriction: Optional[float] = Field(alias="bedFriction")
    subtype: Optional[CulvertSubType] = Field(
        CulvertSubType.culvert.value, alias="subType"
    )
    bendlosscoeff: Optional[float] = Field(alias="bendLossCoeff")

    _split_to_list = get_split_string_on_delimiter_validator("relopening", "losscoeff")
    _flowdirection_validator = get_enum_validator("allowedflowdir", enum=FlowDirection)
    _subtype_validator = get_enum_validator("subtype", enum=CulvertSubType)
    _frictiontype_validator = get_enum_validator("bedfrictiontype", enum=FrictionType)

    @root_validator(allow_reuse=True)
    def validate_that_valve_related_fields_are_present_for_culverts_with_valves(
        cls, values: Dict
    ) -> Dict:
        """Validates that valve-related fields are present when there is a valve present."""
        return validate_required_fields(
            values,
            "valveopeningheight",
            "numlosscoeff",
            "relopening",
            "losscoeff",
            conditional_field_name="valveonoff",
            conditional_value=True,
        )

    @root_validator(allow_reuse=True)
    def validate_that_bendlosscoeff_field_is_present_for_invertedsyphons(
        cls, values: Dict
    ) -> Dict:
        """Validates that the bendlosscoeff value is present when dealing with inverted syphons."""
        return validate_required_fields(
            values,
            "bendlosscoeff",
            conditional_field_name="subtype",
            conditional_value=CulvertSubType.invertedSiphon,
        )

    @root_validator(allow_reuse=True)
    def check_list_lengths(cls, values):
        """Validates that the length of the relopening and losscoeff fields are as expected."""
        return validate_correct_length(
            values,
            "relopening",
            "losscoeff",
            length_name="numlosscoeff",
            list_required_with_length=True,
        )

    @root_validator(allow_reuse=True)
    def validate_that_bendlosscoeff_is_not_provided_for_culverts(
        cls, values: Dict
    ) -> Dict:
        """Validates that the bendlosscoeff field is not provided when the subtype is a culvert."""
        return validate_forbidden_fields(
            values,
            "bendlosscoeff",
            conditional_field_name="subtype",
            conditional_value=CulvertSubType.culvert,
        )


class Pump(Structure):
    """
    Hydraulic structure with `type=pump`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the pump input as described in
    [UM Sec.C.12.6](https://content.oss.deltares.nl/delft3d/manuals/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.6).
    """

    type: Literal["pump"] = Field("pump", alias="type")

    orientation: Optional[Orientation] = Field(alias="orientation")
    controlside: Optional[str] = Field(alias="controlSide")  # TODO Enum
    numstages: Optional[int] = Field(alias="numStages")
    capacity: Union[float, Path] = Field(alias="capacity")

    startlevelsuctionside: Optional[List[float]] = Field(alias="startLevelSuctionSide")
    stoplevelsuctionside: Optional[List[float]] = Field(alias="stopLevelSuctionSide")
    startleveldeliveryside: Optional[List[float]] = Field(
        alias="startLevelDeliverySide"
    )
    stopleveldeliveryside: Optional[List[float]] = Field(alias="stopLevelDeliverySide")
    numreductionlevels: Optional[int] = Field(alias="numReductionLevels")
    head: Optional[List[float]] = Field(alias="head")
    reductionfactor: Optional[List[float]] = Field(alias="reductionFactor")

    _split_to_list = get_split_string_on_delimiter_validator(
        "startlevelsuctionside",
        "stoplevelsuctionside",
        "startleveldeliveryside",
        "stopleveldeliveryside",
        "head",
        "reductionfactor",
    )

    _orientation_validator = get_enum_validator("orientation", enum=Orientation)

    @root_validator(allow_reuse=True)
    def validate_that_controlside_is_provided_when_numstages_is_provided(
        cls, values: Dict
    ) -> Dict:
        return validate_required_fields(
            values,
            "controlside",
            conditional_field_name="numstages",
            conditional_value=0,
            comparison_func=gt,
        )

    @classmethod
    def _check_list_lengths_suctionside(cls, values: Dict) -> Dict:
        """Validates that the length of the startlevelsuctionside and stoplevelsuctionside fields are as expected."""
        return validate_correct_length(
            values,
            "startlevelsuctionside",
            "stoplevelsuctionside",
            length_name="numstages",
            list_required_with_length=True,
        )

    @root_validator(allow_reuse=True)
    def conditionally_check_list_lengths_suctionside(cls, values: Dict) -> Dict:
        """
        Validates the length of the suction side fields, but only if there is a controlside value
        present in the values and the controlside is not equal to the deliverySide.
        """
        return validate_conditionally(
            cls,
            values,
            Pump._check_list_lengths_suctionside,
            "controlside",
            "deliverySide",
            ne,
        )

    @classmethod
    def _check_list_lengths_deliveryside(cls, values: Dict) -> Dict:
        """Validates that the length of the startleveldeliveryside and stopleveldeliveryside fields are as expected."""
        return validate_correct_length(
            values,
            "startleveldeliveryside",
            "stopleveldeliveryside",
            length_name="numstages",
            list_required_with_length=True,
        )

    @root_validator(allow_reuse=True)
    def conditionally_check_list_lengths_deliveryside(cls, values: Dict) -> Dict:
        """
        Validates the length of the delivery side fields, but only if there is a controlside value
        present in the values and the controlside is not equal to the suctionSide.
        """
        return validate_conditionally(
            cls,
            values,
            Pump._check_list_lengths_deliveryside,
            "controlside",
            "suctionSide",
            ne,
        )

    @root_validator(allow_reuse=True)
    def check_list_lengths_head_and_reductionfactor(cls, values):
        """Validates that the lengths of the head and reductionfactor fields are as expected."""
        return validate_correct_length(
            values,
            "head",
            "reductionfactor",
            length_name="numreductionlevels",
            list_required_with_length=True,
        )


class Compound(Structure):
    """
    Hydraulic structure with `type=compound`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the compound input as described in
    [UM Sec.C.12.11](https://content.oss.deltares.nl/delft3d/manuals/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.11).
    """

    type: Literal["compound"] = Field("compound", alias="type")
    numstructures: int = Field(alias="numStructures")
    structureids: List[str] = Field(alias="structureIds", delimiter=";")

    _split_to_list = get_split_string_on_delimiter_validator(
        "structureids",
    )


class Orifice(Structure):
    """
    Hydraulic structure with `type=orifice`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the orifice input as described in
    [UM Sec.C.12.7](https://content.oss.deltares.nl/delft3d/manuals/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.7).
    """

    type: Literal["orifice"] = Field("orifice", alias="type")
    allowedflowdir: Optional[FlowDirection] = Field(
        FlowDirection.both.value, alias="allowedFlowDir"
    )

    crestlevel: Union[float, Path] = Field(alias="crestLevel")
    crestwidth: Optional[float] = Field(None, alias="crestWidth")
    gateloweredgelevel: Union[float, Path] = Field(alias="gateLowerEdgeLevel")
    corrcoeff: float = Field(1.0, alias="corrCoeff")
    usevelocityheight: bool = Field(True, alias="useVelocityHeight")

    # TODO Use a validator here to check the optionals related to the bool field
    uselimitflowpos: Optional[bool] = Field(False, alias="useLimitFlowPos")
    limitflowpos: Optional[float] = Field(alias="limitFlowPos")

    uselimitflowneg: Optional[bool] = Field(False, alias="useLimitFlowNeg")
    limitflowneg: Optional[float] = Field(alias="limitFlowNeg")

    _flowdirection_validator = get_enum_validator("allowedflowdir", enum=FlowDirection)

    @validator("limitflowpos", always=True)
    @classmethod
    def _validate_limitflowpos(cls, v, values):
        return cls._validate_limitflow(v, values, "limitFlowPos", "useLimitFlowPos")

    @validator("limitflowneg", always=True)
    @classmethod
    def _validate_limitflowneg(cls, v, values):
        return cls._validate_limitflow(v, values, "limitFlowNeg", "useLimitFlowNeg")

    @classmethod
    def _validate_limitflow(cls, v, values, limitflow: str, uselimitflow: str):
        if v is None and values[uselimitflow.lower()] == True:
            raise ValueError(
                f"{limitflow} should be defined when {uselimitflow} is true"
            )

        return v


class GateOpeningHorizontalDirection(str, Enum):
    """Horizontal opening direction of gate door[s]."""

    symmetric = "symmetric"
    from_left = "fromLeft"
    from_right = "fromRight"
    allowedvaluestext = "Possible values: symmetric, fromLeft, fromRight."


class GeneralStructure(Structure):
    """
    Hydraulic structure with `type=generalStructure`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the orifice input as described in
    [UM Sec.C.12.9](https://content.oss.deltares.nl/delft3d/manuals/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.9).
    """

    class Comments(Structure.Comments):
        type: Optional[str] = Field(
            "Structure type; must read generalStructure", alias="type"
        )
        allowedflowdir: Optional[str] = Field(
            FlowDirection.allowedvaluestext, alias="allowedFlowDir"
        )

        upstream1width: Optional[str] = Field("w_u1 [m]", alias="upstream1Width")
        upstream1level: Optional[str] = Field("z_u1 [m AD]", alias="upstream1Level")
        upstream2width: Optional[str] = Field("w_u2 [m]", alias="upstream2Width")
        upstream2level: Optional[str] = Field("z_u2 [m D]", alias="upstream2Level")

        crestwidth: Optional[str] = Field("w_s [m]", alias="crestWidth")
        crestlevel: Optional[str] = Field("z_s [m AD]", alias="crestLevel")
        crestlength: Optional[str] = Field(
            "The crest length across the general structure [m]. When the crest length > 0, the extra resistance for this structure will be ls * g/(C2 * waterdepth)",
            alias="crestLength",
        )

        downstream1width: Optional[str] = Field("w_d1 [m]", alias="downstream1Width")
        downstream1level: Optional[str] = Field("z_d1 [m AD]", alias="downstream1Level")
        downstream2width: Optional[str] = Field("w_d2 [m]", alias="downstream2Width")
        downstream2level: Optional[str] = Field("z_d2 [m AD]", alias="downstream2Level")

        gateloweredgelevel: Optional[str] = Field(
            "Position of gate door’s lower edge [m AD]", alias="gateLowerEdgeLevel"
        )
        posfreegateflowcoeff: Optional[str] = Field(
            "Positive free gate flow corr.coeff. cgf [-]", alias="posFreeGateFlowCoeff"
        )
        posdrowngateflowcoeff: Optional[str] = Field(
            "Positive drowned gate flow corr.coeff. cgd [-]",
            alias="posDrownGateFlowCoeff",
        )
        posfreeweirflowcoeff: Optional[str] = Field(
            "Positive free weir flow corr.coeff. cwf [-]", alias="posFreeWeirFlowCoeff"
        )
        posdrownweirflowcoeff: Optional[str] = Field(
            "Positive drowned weir flow corr.coeff. cwd [-]",
            alias="posDrownWeirFlowCoeff",
        )
        poscontrcoeffreegate: Optional[str] = Field(
            "Positive gate flow contraction coefficient µgf [-]",
            alias="posContrCoefFreeGate",
        )
        negfreegateflowcoeff: Optional[str] = Field(
            "Negative free gate flow corr.coeff. cgf [-]", alias="negFreeGateFlowCoeff"
        )
        negdrowngateflowcoeff: Optional[str] = Field(
            "Negative drowned gate flow corr.coeff. cgd [-]",
            alias="negDrownGateFlowCoeff",
        )
        negfreeweirflowcoeff: Optional[str] = Field(
            "Negative free weir flow corr.coeff. cwf [-]", alias="negFreeWeirFlowCoeff"
        )
        negdrownweirflowcoeff: Optional[str] = Field(
            "Negative drowned weir flow corr.coeff. cwd [-]",
            alias="negDrownWeirFlowCoeff",
        )
        negcontrcoeffreegate: Optional[str] = Field(
            "Negative gate flow contraction coefficient mu gf [-]",
            alias="negContrCoefFreeGate",
        )
        extraresistance: Optional[str] = Field(
            "Extra resistance [-]", alias="extraResistance"
        )
        gateheight: Optional[str] = Field(None, alias="gateHeight")
        gateopeningwidth: Optional[str] = Field(
            "Opening width between gate doors [m], should be smaller than (or equal to) crestWidth",
            alias="gateOpeningWidth",
        )
        gateopeninghorizontaldirection: Optional[str] = Field(
            "Horizontal opening direction of gate door[s]. Possible values are: symmetric, fromLeft, fromRight",
            alias="gateOpeningHorizontalDirection",
        )
        usevelocityheight: Optional[str] = Field(
            "Flag indicates whether the velocity height is to be calculated or not",
            alias="useVelocityHeight",
        )

    comments: Optional[Comments] = Comments()

    type: Literal["generalStructure"] = Field("generalStructure", alias="type")
    allowedflowdir: Optional[FlowDirection] = Field(
        FlowDirection.both.value, alias="allowedFlowDir"
    )

    upstream1width: Optional[float] = Field(10.0, alias="upstream1Width")
    upstream1level: Optional[float] = Field(0.0, alias="upstream1Level")
    upstream2width: Optional[float] = Field(10.0, alias="upstream2Width")
    upstream2level: Optional[float] = Field(0.0, alias="upstream2Level")

    crestwidth: Optional[float] = Field(10.0, alias="crestWidth")
    crestlevel: Optional[float] = Field(0.0, alias="crestLevel")
    crestlength: Optional[float] = Field(0.0, alias="crestLength")

    downstream1width: Optional[float] = Field(10.0, alias="downstream1Width")
    downstream1level: Optional[float] = Field(0.0, alias="downstream1Level")
    downstream2width: Optional[float] = Field(10.0, alias="downstream2Width")
    downstream2level: Optional[float] = Field(0.0, alias="downstream2Level")

    gateloweredgelevel: Optional[float] = Field(11.0, alias="gateLowerEdgeLevel")
    posfreegateflowcoeff: Optional[float] = Field(1.0, alias="posFreeGateFlowCoeff")
    posdrowngateflowcoeff: Optional[float] = Field(1.0, alias="posDrownGateFlowCoeff")
    posfreeweirflowcoeff: Optional[float] = Field(1.0, alias="posFreeWeirFlowCoeff")
    posdrownweirflowcoeff: Optional[float] = Field(1.0, alias="posDrownWeirFlowCoeff")
    poscontrcoeffreegate: Optional[float] = Field(1.0, alias="posContrCoefFreeGate")
    negfreegateflowcoeff: Optional[float] = Field(1.0, alias="negFreeGateFlowCoeff")
    negdrowngateflowcoeff: Optional[float] = Field(1.0, alias="negDrownGateFlowCoeff")
    negfreeweirflowcoeff: Optional[float] = Field(1.0, alias="negFreeWeirFlowCoeff")
    negdrownweirflowcoeff: Optional[float] = Field(1.0, alias="negDrownWeirFlowCoeff")
    negcontrcoeffreegate: Optional[float] = Field(1.0, alias="negContrCoefFreeGate")
    extraresistance: Optional[float] = Field(0.0, alias="extraResistance")
    gateheight: Optional[float] = Field(1e10, alias="gateHeight")
    gateopeningwidth: Optional[float] = Field(0.0, alias="gateOpeningWidth")
    gateopeninghorizontaldirection: Optional[GateOpeningHorizontalDirection] = Field(
        GateOpeningHorizontalDirection.symmetric.value,
        alias="gateOpeningHorizontalDirection",
    )
    usevelocityheight: Optional[bool] = Field(True, alias="useVelocityHeight")


class DambreakAlgorithm(int, Enum):
    van_der_knaap = 1  # "van der Knaap, 2000"
    verheij_van_der_knaap = 2  # "Verheij-van der Knaap, 2002"
    timeseries = 3  # "Predefined time series, dambreakLevelsAndWidths."

    @property
    def description(self) -> str:
        """
        Property to return the description of the enums defined above.
        Useful for comments in output files.

        Returns:
            str: Description for the current enum.
        """
        description_dict = dict(
            van_der_knaap="van der Knaap, 2000",
            verheij_van_der_knaap="Verheij-van der Knaap, 2002",
            timeseries="Predefined time series, dambreakLevelsAndWidths",
        )
        return description_dict[self.name]


class Dambreak(Structure):
    """
    Hydraulic structure with `type=dambreak`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the dambreak input as described in
    [UM Sec.C.12.10](https://content.oss.deltares.nl/delft3d/manuals/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.10).
    """

    class Comments(Structure.Comments):
        type: Optional[str] = Field("Structure type; must read dambreak", alias="type")
        startlocationx: Optional[str] = Field(
            "x-coordinate of breach starting point.", alias="startLocationX"
        )
        startlocationy: Optional[str] = Field(
            "y-coordinate of breach starting point.", alias="startLocationY"
        )
        algorithm: Optional[str] = Field(
            "Breach growth algorithm. Possible values are: 1 (van der Knaap (2000)), 2 (Verheij–van der Knaap (2002)), 3: Predefined time series, see dambreakLevelsAndWidths",
            alias="algorithm",
        )
        crestlevelini: Optional[str] = Field(
            "Initial breach level zcrest level [m AD].", alias="crestLevelIni"
        )
        breachwidthini: Optional[str] = Field(
            "Initial breach width B0 [m].", alias="breachWidthIni"
        )
        crestlevelmin: Optional[str] = Field(
            "Minimal breach level zmin [m AD].", alias="crestLevelMin"
        )
        t0: Optional[str] = Field("Breach start time Tstart [s].", alias="t0")
        timetobreachtomaximumdepth: Optional[str] = Field(
            "tPhase 1 [s].", alias="timeToBreachToMaximumDepth"
        )
        f1: Optional[str] = Field("Factor f1 [-]", alias="f1")
        f2: Optional[str] = Field("Factor f2 [-]", alias="f2")
        ucrit: Optional[str] = Field(
            "Critical flow velocity uc for erosion [m/s].", alias="uCrit"
        )
        waterlevelupstreamlocationx: Optional[str] = Field(
            "(optional) x-coordinate of custom upstream water level point.",
            alias="waterLevelUpstreamLocationX",
        )
        waterlevelupstreamlocationy: Optional[str] = Field(
            "(optional) y-coordinate of custom upstream water level point.",
            alias="waterLevelUpstreamLocationY",
        )
        waterleveldownstreamlocationx: Optional[str] = Field(
            "(optional) x-coordinate of custom downstream water level point.",
            alias="waterLevelDownstreamLocationX",
        )
        waterleveldownstreamlocationy: Optional[str] = Field(
            "(optional) y-coordinate of custom downstream water level point.",
            alias="waterLevelDownstreamLocationY",
        )
        waterlevelupstreamnodeid: Optional[str] = Field(
            "(optional) Node Id of custom upstream water level point.",
            alias="waterLevelUpstreamNodeId",
        )
        waterleveldownstreamnodeid: Optional[str] = Field(
            "(optional) Node Id of custom downstream water level point.",
            alias="waterLevelDownstreamNodeId",
        )
        dambreaklevelsandwidths: Optional[str] = Field(
            "(only when algorithm=3) Filename of <*.tim> file (Section C.4) containing the breach levels and widths.",
            alias="dambreakLevelsAndWidths",
        )

    comments: Comments = Comments()
    type: Literal["dambreak"] = Field("dambreak", alias="type")
    startlocationx: float = Field(alias="startLocationX")
    startlocationy: float = Field(alias="startLocationY")
    algorithm: DambreakAlgorithm = Field(alias="algorithm")

    crestlevelini: float = Field(alias="crestLevelIni")
    breachwidthini: float = Field(alias="breachWidthIni")
    crestlevelmin: float = Field(alias="crestLevelMin")
    t0: float = Field(alias="t0")
    timetobreachtomaximumdepth: float = Field(alias="timeToBreachToMaximumDepth")
    f1: float = Field(alias="f1")
    f2: float = Field(alias="f2")
    ucrit: float = Field(alias="uCrit")
    waterlevelupstreamlocationx: Optional[float] = Field(
        alias="waterLevelUpstreamLocationX"
    )
    waterlevelupstreamlocationy: Optional[float] = Field(
        alias="waterLevelUpstreamLocationY"
    )
    waterleveldownstreamlocationx: Optional[float] = Field(
        alias="waterLevelDownstreamLocationX"
    )
    waterleveldownstreamlocationy: Optional[float] = Field(
        alias="waterLevelDownstreamLocationY"
    )
    waterlevelupstreamnodeid: Optional[str] = Field(alias="waterLevelUpstreamNodeId")
    waterleveldownstreamnodeid: Optional[str] = Field(
        alias="waterLevelDownstreamNodeId"
    )
    dambreaklevelsandwidths: Optional[Path] = Field(alias="dambreakLevelsAndWidths")

    @validator("algorithm", pre=True)
    @classmethod
    def validate_algorithm(cls, value: str) -> DambreakAlgorithm:
        """
        Validates the algorithm parameter for the dambreak structure.

        Args:
            value (int): algorithm value read from the user's input.

        Raises:
            ValueError: When the value given is not of type int.
            ValueError: When the value given is not in the range [1,3]

        Returns:
            int: Validated value.
        """
        int_value = -1
        try:
            int_value = int(value)
        except Exception:
            raise ValueError("Dambreak algorithm value should be of type int.")
        if 0 < int_value <= 3:
            return DambreakAlgorithm(int_value)
        raise ValueError("Dambreak algorithm value should be 1, 2 or 3.")

    @validator("dambreaklevelsandwidths")
    @classmethod
    def validate_dambreak_levels_and_widths(
        cls, field_value: Optional[Path], values: dict
    ) -> Optional[Path]:
        """
        Validates whether a dambreak can be created with the given dambreakLevelsAndWidths
        property. This property should be given when the algorithm value is 3.

        Args:
            field_value (Optional[Path]): Value given for dambreakLevelsAndWidths.
            values (dict): Dictionary of values already validated (assuming algorithm is in it).

        Raises:
            ValueError: When algorithm value is not 3 and field_value has a value.

        Returns:
            Optional[Path]: The value given for dambreakLevelsAndwidths.
        """
        # Retrieve the algorithm value (if not found use 0).
        algorithm_value = values.get("algorithm", 0)
        if field_value is not None and algorithm_value != 3:
            # dambreakLevelsAndWidths can only be set when algorithm = 3
            raise ValueError(
                f"Dambreak field dambreakLevelsAndWidths can only be set when algorithm = 3, current value: {algorithm_value}."
            )
        return field_value

    @root_validator
    @classmethod
    def check_location_dambreak(cls, values: dict) -> dict:
        """
        Verifies whether the location for this structure contains valid values for
        numCoordinates, xCoordinates and yCoordinates or instead is using a polyline file.
        Verifies whether de water level location specifications are valid.

        Args:
            values (dict): Dictionary of validated values to create a Dambreak.

        Raises:
            ValueError: When the values dictionary does not contain valid coordinates or polyline file or when the water level location specifications are not valid.

        Returns:
            dict: Dictionary of validated values.
        """

        def _validate_waterlevel_location(x_key: str, y_key: str, node_key: str):
            x_is_given = values.get(x_key.lower()) is not None
            y_is_given = values.get(y_key.lower()) is not None
            node_is_given = values.get(node_key.lower()) is not None

            if (x_is_given and y_is_given and not node_is_given) or (
                node_is_given and not x_is_given and not y_is_given
            ):
                return

            raise ValueError(
                f"Either `{node_key}` should be specified or `{x_key}` and `{y_key}`."
            )

        _validate_waterlevel_location(
            "waterLevelUpstreamLocationX",
            "waterLevelUpstreamLocationY",
            "waterLevelUpstreamNodeId",
        )
        _validate_waterlevel_location(
            "waterLevelDownstreamLocationX",
            "waterLevelDownstreamLocationY",
            "waterLevelDownstreamNodeId",
        )

        return values


class Bridge(Structure):
    """
    Hydraulic structure with `type=bridge`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the bridge input as described in
    [UM Sec.C.12.5](https://content.oss.deltares.nl/delft3d/manuals/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.5).
    """

    class Comments(Structure.Comments):
        type: Optional[str] = Field("Structure type; must read bridge", alias="type")
        allowedflowdir: Optional[str] = Field(
            FlowDirection.allowedvaluestext, alias="allowedFlowdir"
        )

        csdefid: Optional[str] = Field(
            "Id of Cross-Section Definition.", alias="csDefId"
        )
        shift: Optional[str] = Field(
            "Vertical shift of the cross section definition [m]. Defined positive upwards."
        )
        inletlosscoeff: Optional[str] = Field(
            "Inlet loss coefficient [-], ξ_i.",
            alias="inletLossCoeff",
        )
        outletlosscoeff: Optional[str] = Field(
            "Outlet loss coefficient [-], k.",
            alias="outletLossCoeff",
        )
        frictiontype: Optional[str] = Field(
            "Friction type, possible values are: Chezy, Manning, wallLawNikuradse, WhiteColebrook, StricklerNikuradse, Strickler, deBosBijkerk.",
            alias="frictionType",
        )
        friction: Optional[str] = Field(
            "Friction value, used in friction loss.",
            alias="friction",
        )
        length: Optional[str] = Field("Length [m], L.")

    comments: Comments = Comments()

    type: Literal["bridge"] = Field("bridge", alias="type")
    allowedflowdir: FlowDirection = Field(alias="allowedFlowdir")

    csdefid: str = Field(alias="csDefId")
    shift: float
    inletlosscoeff: float = Field(alias="inletLossCoeff")
    outletlosscoeff: float = Field(alias="outletLossCoeff")
    frictiontype: FrictionType = Field(alias="frictionType")
    friction: float
    length: float

    _frictiontype_validator = get_enum_validator("frictiontype", enum=FrictionType)


class StructureGeneral(INIGeneral):
    """`[General]` section with structure file metadata."""

    _header: Literal["General"] = "General"
    fileversion: str = Field("3.00", alias="fileVersion")
    filetype: Literal["structure"] = Field("structure", alias="fileType")


class StructureModel(INIModel):
    """
    The overall structure model that contains the contents of one structure file.

    This model is typically referenced under a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[..]`.

    Attributes:
        general (StructureGeneral): `[General]` block with file metadata.
        branch (List[Structure]): List of `[Structure]` blocks for all hydraulic structures.
    """

    general: StructureGeneral = StructureGeneral()
    structure: List[Structure] = []

    @classmethod
    def _ext(cls) -> str:
        return ".ini"

    @classmethod
    def _filename(cls) -> str:
        return "structures"

    _split_to_list = make_list_validator("structure")



from .models import (
    AveragingType,
    DataFileType,
    IniFieldGeneral,
    IniFieldModel,
    InitialField,
    InterpolationMethod,
    Operand,
    ParameterField,
)

__all__ = [
    "DataFileType",
    "InterpolationMethod",
    "Operand",
    "AveragingType",
    "IniFieldGeneral",
    "InitialField",
    "ParameterField",
    "IniFieldModel",
]



from .bc import *
from .common import *
from .crosssection import *
from .ext import *
from .friction import *
from .gui import *
from .inifield import *
from .mdu import *
from .net import *
from .obs import *
from .obscrosssection import *
from .onedfield import *
from .polyfile import *
from .storagenode import *
from .structure import *
from .xyz import *



from .models import FrictBranch, FrictGeneral, FrictGlobal, FrictionModel, FrictionType

__all__ = [
    "FrictionType",
    "FrictGeneral",
    "FrictGlobal",
    "FrictBranch",
    "FrictionModel",
]



from itertools import chain
from pathlib import Path
from typing import Generator, Iterable, Optional, Sequence

from hydrolib.core.basemodel import SerializerConfig
from hydrolib.core.dflowfm.polyfile.models import (
    Description,
    Metadata,
    Point,
    PolyObject,
)


class Serializer:
    """Serializer provides several static serialize methods for the models."""

    @staticmethod
    def serialize_description(description: Optional[Description]) -> Iterable[str]:
        """Serialize the Description to a string which can be used within a polyfile.

        Returns:
            str: The serialised equivalent of this Description
        """
        if description is None:
            return []
        if description.content == "":
            return [
                "*",
            ]
        return (f"*{v.rstrip()}" for v in description.content.splitlines())

    @staticmethod
    def serialize_metadata(metadata: Metadata) -> Iterable[str]:
        """Serialize this Metadata to a string which can be used within a polyfile.

        The number of rows and number of columns are separated by four spaces.

        Returns:
            str: The serialised equivalent of this Metadata
        """
        return [metadata.name, f"{metadata.n_rows}    {metadata.n_columns}"]

    @staticmethod
    def serialize_point(point: Point, config: SerializerConfig) -> str:
        """Serialize this Point to a string which can be used within a polyfile.

        the point data is indented with 4 spaces, and the individual values are
        separated by 4 spaces as well.

        Args:
            point (Point): The point to serialize.
            config (SerializerConfig): The serialization configuration.

        Returns:
            str: The serialised equivalent of this Point
        """
        space = 4 * " "
        float_format = lambda v: f"{v:{config.float_format}}"
        return space + space.join(
            float_format(v) for v in Serializer._get_point_values(point)
        )

    @staticmethod
    def _get_point_values(point: Point) -> Generator[float, None, None]:
        yield point.x
        yield point.y
        if point.z:
            yield point.z
        for value in point.data:
            yield value

    @staticmethod
    def serialize_poly_object(
        obj: PolyObject, config: SerializerConfig
    ) -> Iterable[str]:
        """Serialize this PolyObject to a string which can be used within a polyfile.

        Args:
            obj (PolyObject): The poly object to serializer.
            config (SerializerConfig): The serialization configuration.

        Returns:
            str: The serialised equivalent of this PolyObject
        """

        description = Serializer.serialize_description(obj.description)
        metadata = Serializer.serialize_metadata(obj.metadata)
        points = [Serializer.serialize_point(obj, config) for obj in obj.points]
        return chain(description, metadata, points)


def write_polyfile(
    path: Path, data: Sequence[PolyObject], config: SerializerConfig
) -> None:
    """Write the data to a new file at path

    Args:
        path (Path): The path to write the data to
        data (Sequence[PolyObject]): The poly objects to write
        config (SerializerConfig): The serialization configuration.
    """
    serialized_poly_objects = [
        Serializer.serialize_poly_object(poly_object, config) for poly_object in data
    ]
    serialized_data = chain.from_iterable(serialized_poly_objects)

    path.parent.mkdir(parents=True, exist_ok=True)

    with path.open("w") as f:

        for line in serialized_data:
            f.write(line)
            f.write("\n")



from .models import (
    AutoStartOption,
    Calibration,
    ExternalForcing,
    FMModel,
    General,
    Geometry,
    GroundWater,
    Hydrology,
    InfiltrationMethod,
    Numerics,
    Output,
    Particles,
    ParticlesThreeDType,
    Physics,
    Processes,
    ProcessFluxIntegration,
    Restart,
    Sediment,
    Time,
    Trachytopes,
    Vegetation,
    VegetationModelNr,
    VolumeTables,
    Waves,
    Wind,
)

__all__ = [
    "AutoStartOption",
    "General",
    "Numerics",
    "VolumeTables",
    "Physics",
    "Sediment",
    "Wind",
    "Waves",
    "Time",
    "Restart",
    "ExternalForcing",
    "Hydrology",
    "Trachytopes",
    "Output",
    "Geometry",
    "Calibration",
    "InfiltrationMethod",
    "GroundWater",
    "ProcessFluxIntegration",
    "Processes",
    "ParticlesThreeDType",
    "Particles",
    "VegetationModelNr",
    "Vegetation",
    "FMModel",
]



from .models import (
    ObservationCrossSection,
    ObservationCrossSectionGeneral,
    ObservationCrossSectionModel,
)

__all__ = [
    "ObservationCrossSectionGeneral",
    "ObservationCrossSection",
    "ObservationCrossSectionModel",
]



from .models import (
    CircleCrsDef,
    CrossDefGeneral,
    CrossDefModel,
    CrossLocGeneral,
    CrossLocModel,
    CrossSection,
    CrossSectionDefinition,
    RectangleCrsDef,
    XYZCrsDef,
    YZCrsDef,
    ZWCrsDef,
    ZWRiverCrsDef,
)

__all__ = [
    "CrossDefGeneral",
    "CrossLocGeneral",
    "CrossSectionDefinition",
    "CrossDefModel",
    "CircleCrsDef",
    "RectangleCrsDef",
    "ZWRiverCrsDef",
    "ZWCrsDef",
    "YZCrsDef",
    "XYZCrsDef",
    "CrossSection",
    "CrossLocModel",
]



import logging
from abc import ABC
from enum import Enum
from typing import Any, Callable, List, Literal, Optional, Set, Type, Union

from pydantic import Extra, Field, root_validator
from pydantic.class_validators import validator
from pydantic.fields import ModelField

from hydrolib.core import __version__ as version
from hydrolib.core.basemodel import BaseModel, ParsableFileModel

from ..ini.io_models import CommentBlock, Document, Property, Section
from .parser import Parser
from .serializer import (
    DataBlockINIBasedSerializerConfig,
    INISerializerConfig,
    write_ini,
)
from .util import make_list_validator

logger = logging.getLogger(__name__)


class INIBasedModel(BaseModel, ABC):
    """INIBasedModel defines the base model for blocks/chapters
    inside an INIModel (*.ini file).

    INIBasedModel instances can be created from Section instances
    obtained through parsing ini documents. It further supports
    adding arbitrary fields to it, which will be written to file.
    Lastly, no arbitrary types are allowed for the defined fields.

    Attributes:
        comments (Optional[Comments]):
            Optional Comments if defined by the user, containing
            descriptions for all data fields.
    """

    _header: str

    class Config:
        extra = Extra.ignore
        arbitrary_types_allowed = False

    @classmethod
    def _supports_comments(cls):
        return True

    @classmethod
    def _duplicate_keys_as_list(cls):
        return False

    @classmethod
    def get_list_delimiter(cls) -> str:
        """List delimiter string that will be used for serializing
        list field values for any IniBasedModel, **if** that field has
        no custom list delimiter.

        This function should be overridden by any subclass for a particular
        filetype that needs a specific/different list separator.
        """
        return " "

    @classmethod
    def get_list_field_delimiter(cls, field_key: str) -> str:
        """List delimiter string that will be used for serializing
        the given field's value.
        The returned delimiter is either the field's custom list delimiter
        if that was specified using Field(.., delimiter=".."), or the
        default list delimiter for the model class that this field belongs
        to.

        Args:
            field_key (str): the original field key (not its alias).
        """
        delimiter = None
        if (field := cls.__fields__.get(field_key)) and isinstance(field, ModelField):
            delimiter = field.field_info.extra.get("delimiter")
        if not delimiter:
            delimiter = cls.get_list_delimiter()

        return delimiter

    class Comments(BaseModel, ABC):
        """Comments defines the comments of an INIBasedModel"""

        class Config:
            extra = Extra.allow
            arbitrary_types_allowed = False

    comments: Optional[Comments] = Comments()

    @root_validator(pre=True)
    def _skip_nones_and_set_header(cls, values):
        """Drop None fields for known fields."""
        dropkeys = []
        for k, v in values.items():
            if v is None and k in cls.__fields__.keys():
                dropkeys.append(k)

        logger.info(f"Dropped unset keys: {dropkeys}")
        for k in dropkeys:
            values.pop(k)

        if "_header" in values:
            values["_header"] = cls._header

        return values

    @validator("comments", always=True, allow_reuse=True)
    def comments_matches_has_comments(cls, v):
        if not cls._supports_comments() and v is not None:
            logging.warning(f"Dropped unsupported comments from {cls.__name__} init.")
            v = None
        return v

    @classmethod
    def validate(cls: Type["INIBasedModel"], value: Any) -> "INIBasedModel":
        if isinstance(value, Section):
            value = value.flatten(
                cls._duplicate_keys_as_list(), cls._supports_comments()
            )

        return super().validate(value)

    @classmethod
    def _exclude_fields(cls) -> Set:
        return {"comments", "datablock", "_header"}

    @classmethod
    def _convert_value(cls, key: str, v: Any, config: INISerializerConfig) -> str:
        if isinstance(v, bool):
            return str(int(v))
        elif isinstance(v, list):
            float_format = (
                lambda x: f"{x:{config.float_format}}"
                if isinstance(x, float)
                else str(x)
            )
            return cls.get_list_field_delimiter(key).join([float_format(x) for x in v])
        elif isinstance(v, Enum):
            return v.value
        elif isinstance(v, float):
            return f"{v:{config.float_format}}"
        elif v is None:
            return ""
        else:
            return str(v)

    def _to_section(self, config: INISerializerConfig) -> Section:
        props = []
        for key, value in self:
            if key in self._exclude_fields():
                continue

            field_key = key
            if key in self.__fields__:
                key = self.__fields__[key].alias

            prop = Property(
                key=key,
                value=self.__class__._convert_value(field_key, value, config),
                comment=getattr(self.comments, key.lower(), None),
            )
            props.append(prop)
        return Section(header=self._header, content=props)


class DataBlockINIBasedModel(INIBasedModel):
    """DataBlockINIBasedModel defines the base model for ini models with datablocks.

    Attributes:
        datablock (List[List[Union[float, str]]]): (class attribute) the actual data
            columns.
    """

    datablock: List[List[Union[float, str]]] = []

    _make_lists = make_list_validator("datablock")

    def _to_section(self, config: DataBlockINIBasedSerializerConfig) -> Section:
        section = super()._to_section(config)
        section.datablock = self._to_datablock(config)
        return section

    def _to_datablock(self, config: DataBlockINIBasedSerializerConfig) -> List[List]:
        converted_datablock = []

        for row in self.datablock:
            converted_row = (
                DataBlockINIBasedModel.convert_value(value, config) for value in row
            )
            converted_datablock.append(list(converted_row))

        return converted_datablock

    @classmethod
    def convert_value(
        cls, value: Union[float, str], config: DataBlockINIBasedSerializerConfig
    ) -> str:
        if isinstance(value, float):
            return f"{value:{config.float_format_datablock}}"

        return value


class INIGeneral(INIBasedModel):
    _header: Literal["General"] = "General"
    fileversion: str = Field("3.00", alias="fileVersion")
    filetype: str = Field(alias="fileType")

    @classmethod
    def _supports_comments(cls):
        return True


class INIModel(ParsableFileModel):
    """INI Model representation of a *.ini file.

    Typically subclasses will implement the various sorts of ini files,
    specifically for their fileType/contents.
    Child elements of this class associated with chapters/blocks in the
    ini file will be (sub)class of INIBasedModel.
    """

    serializer_config: INISerializerConfig = INISerializerConfig()

    general: INIGeneral

    @classmethod
    def _ext(cls) -> str:
        return ".ini"

    @classmethod
    def _filename(cls) -> str:
        return "fm"

    @classmethod
    def _get_serializer(cls):
        pass  # unused in favor of direct _serialize

    @classmethod
    def _get_parser(cls) -> Callable:
        return Parser.parse_as_dict

    def _to_document(self) -> Document:
        header = CommentBlock(lines=[f"written by HYDROLIB-core {version}"])
        sections = []
        for key, value in self:
            if key in self._exclude_fields() or value is None:
                continue
            if isinstance(value, list):
                for v in value:
                    sections.append(v._to_section(self.serializer_config))
            else:
                sections.append(value._to_section(self.serializer_config))
        return Document(header_comment=[header], sections=sections)

    def _serialize(self, _: dict) -> None:
        write_ini(
            self._resolved_filepath, self._to_document(), config=self.serializer_config
        )



from typing import Dict, List, Literal, Optional

from pydantic.class_validators import root_validator
from pydantic.fields import Field

from hydrolib.core.dflowfm.common.models import LocationType
from hydrolib.core.dflowfm.ini.models import INIBasedModel, INIGeneral, INIModel
from hydrolib.core.dflowfm.ini.util import (
    LocationValidationConfiguration,
    LocationValidationFieldNames,
    get_enum_validator,
    make_list_validator,
    validate_location_specification,
)


class ObservationPointGeneral(INIGeneral):
    """The observation point file's `[General]` section with file meta data."""

    class Comments(INIBasedModel.Comments):
        fileversion: Optional[str] = Field(
            "File version. Do not edit this.", alias="fileVersion"
        )
        filetype: Optional[str] = Field(
            "File type. Should be 'obsPoints'. Do not edit this.",
            alias="fileType",
        )

    comments: Comments = Comments()
    _header: Literal["General"] = "General"
    fileversion: str = Field("2.00", alias="fileVersion")
    filetype: Literal["obsPoints"] = Field("obsPoints", alias="fileType")


class ObservationPoint(INIBasedModel):
    """
    An observation point that is included in the observation point file.

    All lowercased attributes match with the observation point input as described in
    [UM Sec.F.2.2.1](https://content.oss.deltares.nl/delft3d/manuals/D-Flow_FM_User_Manual_1D2D.pdf#section.F.2.2.1).
    """

    class Comments(INIBasedModel.Comments):
        name: Optional[str] = "Name of the observation point (max. 255 characters)."
        locationtype: Optional[
            str
        ] = "Only when x and y are also specified. 1d: snap to closest 1D grid point, 2d: snap to closest 2D grid cell centre, all: snap to closest 1D or 2D point."
        branchid: Optional[str] = Field(
            "Branch on which the observation point is located.", alias="branchId"
        )
        chainage: Optional[str] = "Chainage on the branch (m)."

        x: Optional[str] = Field(
            "x-coordinate of the location of the observation point.",
            alias="x",
        )
        y: Optional[str] = Field(
            "y-coordinate of the location of the observation point.",
            alias="y",
        )

    comments: Comments = Comments()

    _header: Literal["ObservationPoint"] = "ObservationPoint"

    name: str = Field("id", max_length=255, alias="name")
    locationtype: Optional[LocationType] = Field(None, alias="locationType")

    branchid: Optional[str] = Field(None, alias="branchId")
    chainage: Optional[float] = Field(None, alias="chainage")

    x: Optional[float] = Field(None, alias="x")
    y: Optional[float] = Field(None, alias="y")

    _type_validator = get_enum_validator("locationtype", enum=LocationType)

    @root_validator(allow_reuse=True)
    def validate_that_location_specification_is_correct(cls, values: Dict) -> Dict:
        """Validates that the correct location specification is given."""
        return validate_location_specification(
            values,
            config=LocationValidationConfiguration(
                validate_node=False, validate_num_coordinates=False
            ),
            fields=LocationValidationFieldNames(x_coordinates="x", y_coordinates="y"),
        )

    def _get_identifier(self, data: dict) -> Optional[str]:
        return data.get("name")


class ObservationPointModel(INIModel):
    """
    The overall observation point model that contains the contents of one observation point file.

    This model is typically referenced under a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.output.obsfile[..]`.

    Attributes:
        general (ObservationPointGeneral): `[General]` block with file metadata.
        observationpoint (List[ObservationPoint]): List of `[ObservationPoint]` blocks for all observation points.
    """

    general: ObservationPointGeneral = ObservationPointGeneral()
    observationpoint: List[ObservationPoint] = []

    _make_list = make_list_validator("observationpoint")

    @classmethod
    def _filename(cls) -> str:
        return "obsFile"



from enum import Enum


class LocationType(str, Enum):
    """
    Enum class containing the valid values for the locationType
    attribute in several classes such as Lateral and ObservationPoint.
    """

    oned = "1d"
    """str: Denotes 1D locations (typically 1D pressure points) in a model."""

    twod = "2d"
    """str: Denotes 2D locations (typically 2D grid cells) in a model."""

    all = "all"
    """str: Denotes that both 1D and 2D locations may be selected."""



from .models import (
    Interpolation,
    NodeType,
    StorageNode,
    StorageNodeGeneral,
    StorageNodeModel,
    StorageType,
)

__all__ = [
    "NodeType",
    "StorageType",
    "Interpolation",
    "StorageNodeGeneral",
    "StorageNode",
    "StorageNodeModel",
]



from pathlib import Path
from typing import Callable, Dict, List, Optional

from pydantic.class_validators import root_validator
from pydantic.fields import Field

from hydrolib.core.basemodel import BaseModel, ParsableFileModel, SerializerConfig
from hydrolib.core.rr.topology.parser import NetworkTopologyFileParser
from hydrolib.core.rr.topology.serializer import LinkFileSerializer, NodeFileSerializer

nodetypes_netter_to_rr = {
    43: 1,  # 1: Paved area
    44: 2,  # 2: Unpaved area
    45: 3,  # 3: Greenhouse
    46: 4,  # 4: Open water
    -5: 5,  # 5: Internally reserved for all structures
    34: 6,  # 6: Boundary
    35: 6,  # 6: Boundary
    47: 6,  # 6: Boundary
    48: 8,  # 8: pump
    49: 9,  # 9: weir
    50: 10,  # 10: orifice
    51: 11,  # 11: Manning resistance
    52: 12,  # 12: Q-h relation
    56: 14,  # 14: WWTP (RWZI)
    55: 15,  # 15: Industry
    54: 16,  # 16: Sacramento (ObId ’3B_SACRAMENTO’)
    -21: 21,  # 21: Open water with only precipitation and evaporation (SOBEK3)
    69: 23,  # 23: Wagmod/Walrus
}
""" Dictionary with `nt` mapped against the expected `mt`.

Some model types `mt` do not have a related netter type; in that case the
dict key is a dummy value of -<mt>."""


class Node(BaseModel):
    """Represents a node from the topology node file."""

    id: str = Field(alias="id")
    name: Optional[str] = Field(alias="nm")
    branchid: int = Field(alias="ri")
    modelnodetype: int = Field(alias="mt")
    netternodetype: int = Field(alias="nt")
    objectid: str = Field(alias="ObID")
    xposition: float = Field(alias="px")
    yposition: float = Field(alias="py")

    def _get_identifier(self, data: dict) -> Optional[str]:
        return data.get("id") or data.get("nm")

    def dict(self, *args, **kwargs):
        kwargs["by_alias"] = True
        return super().dict(*args, **kwargs)

    @root_validator()
    @classmethod
    def _validate_node_type(cls, values):

        cls._raise_if_invalid_type(
            values,
            "modelnodetype",
            set(nodetypes_netter_to_rr.values()),
            "model node type (mt)",
        )

        modelnodetype = values.get("modelnodetype")

        # modelnodetype=6 ("boundary node") is a special case that allows various netter nodetypes,
        # so therefore it always validates well.
        if modelnodetype == 6:
            return values

        cls._raise_if_invalid_type(
            values,
            "netternodetype",
            set(nodetypes_netter_to_rr.keys()),
            "netter node type (nt)",
        )

        netternodetype = values.get("netternodetype")
        modelnodetype_expected = nodetypes_netter_to_rr[netternodetype]

        if modelnodetype != modelnodetype_expected:
            raise ValueError(
                f"{modelnodetype} is not a supported model node type (mt) when netter node type (nt) is {netternodetype}. Supported value: {modelnodetype_expected}."
            )

        return values

    @classmethod
    def _raise_if_invalid_type(
        cls, values, field_name: str, supported_values: set, description: str
    ):
        """Validates the node type for the provided `field_name`.
        The specified node type should contain a supported value,
        otherwise a `ValueError` is raised.


        Args:
            values ([type]): Dictionary with values that are used to create this `Node`.
            field_name (str): Field name of the node type to validate.
            supported_values (set): Set of all the supported values for this node type.
            description (str): Description of this node type that will be readable for the user.

        Raises:
            ValueError: Thrown when `supported_values` does node contain the node type.
        """
        field_value = values.get(field_name)

        if field_value not in supported_values:
            str_supported_values = ", ".join([str(t) for t in supported_values])
            raise ValueError(
                f"{field_value} is not a supported {description}. Supported values: {str_supported_values}."
            )


class NodeFile(ParsableFileModel):
    """Represents the file with the RR node topology data."""

    _parser = NetworkTopologyFileParser(enclosing_tag="node")
    node: List[Node] = Field([], alias="node")

    @classmethod
    def _ext(cls) -> str:
        return ".tp"

    @classmethod
    def _filename(cls) -> str:
        return "3b_nod"

    @classmethod
    def _get_serializer(cls) -> Callable[[Path, Dict, SerializerConfig], None]:
        return NodeFileSerializer.serialize

    @classmethod
    def _get_parser(cls) -> Callable:
        return cls._parser.parse


class Link(BaseModel):
    """Represents a link from the topology link file."""

    id: str = Field(alias="id")
    name: Optional[str] = Field(alias="nm")
    branchid: int = Field(alias="ri")
    modellinktype: int = Field(alias="mt")
    branchtype: int = Field(alias="bt")
    objectid: str = Field(alias="ObID")
    beginnode: str = Field(alias="bn")
    endnode: str = Field(alias="en")

    def _get_identifier(self, data: dict) -> Optional[str]:
        return data.get("id") or data.get("nm")

    def dict(self, *args, **kwargs):
        kwargs["by_alias"] = True
        return super().dict(*args, **kwargs)


class LinkFile(ParsableFileModel):
    """Represents the file with the RR link topology data."""

    _parser = NetworkTopologyFileParser(enclosing_tag="brch")
    link: List[Link] = Field([], alias="brch")

    @classmethod
    def _ext(cls) -> str:
        return ".tp"

    @classmethod
    def _filename(cls) -> str:
        return "3b_link"

    @classmethod
    def _get_serializer(cls) -> Callable[[Path, Dict, SerializerConfig], None]:
        return LinkFileSerializer.serialize

    @classmethod
    def _get_parser(cls) -> Callable:
        return cls._parser.parse



"""models.py defines the RainfallRunoffModel and supporting structures."""

from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Optional

from pydantic import Field, validator
from pydantic.types import FilePath

from hydrolib.core.basemodel import (
    DiskOnlyFileModel,
    ParsableFileModel,
    SerializerConfig,
    validator_set_default_disk_only_file_model_when_none,
)

from .meteo.models import BuiModel
from .parser import read
from .serializer import write
from .topology.models import LinkFile, NodeFile


class ImmutableDiskOnlyFileModel(DiskOnlyFileModel, allow_mutation=False):
    """
    ImmutableDiskOnlyFileModel modifies the DiskOnlyFileModel to provide faux
    immutablitity.

    This behaviour is required for the mappix properties, which should always
    have the same name and path and should not be modified by users.
    """

    pass


_mappix_paved_area_sewage_storage_name = "pvstordt.his"
_mappix_paved_area_flow_rates_name = "pvflowdt.his"
_mappix_unpaved_area_flow_rates_name = "upflowdt.his"
_mappix_ground_water_levels_name = "upgwlvdt.his"
_mappix_green_house_bassins_storage_name = "grnstrdt.his"
_mappix_green_house_bassins_results_name = "grnflodt.his"
_mappix_open_water_details_name = "ow_lvldt.his"
_mappix_exceedance_time_reference_levels_name = "ow_excdt.his"
_mappix_flow_rates_over_structures_name = "strflodt.his"
_mappix_flow_rates_to_edge_name = "bndflodt.his"
_mappix_pluvius_max_sewage_storage_name = "plvstrdt.his"
_mappix_pluvius_max_flow_rates_name = "plvflodt.his"
_mappix_balance_name = "balansdt.his"
_mappix_cumulative_balance_name = "cumbaldt.his"
_mappix_salt_concentrations_name = "saltdt.his"


def _validator_mappix_value(field_name: str, expected: str) -> classmethod:
    def validate(v: Any) -> Any:
        if v != expected:
            actual_value = v or "None"
            raise ValueError(
                f"Expected a value of '{expected}' for '{field_name}' but got '{actual_value}'. Mappix values should not be changed."
            )
        return v

    return validator(field_name, allow_reuse=True, pre=True)(validate)


class RainfallRunoffModel(ParsableFileModel):
    """The RainfallRunoffModel contains all paths and sub-models related to the
    Rainfall Runoff model.
    """

    _disk_only_file_model_cannot_be_none = (
        validator_set_default_disk_only_file_model_when_none()
    )

    # Note that order is defined by the .fnm file type and is used for parsing the data.
    control_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("delft_3b.ini"))
    )
    node_data: Optional[NodeFile] = Field(None)
    link_data: Optional[LinkFile] = Field(None)
    open_water_data: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("3brunoff.tp"))
    )
    paved_area_general: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("paved.3b"))
    )
    paved_area_storage: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("paved.sto"))
    )
    paved_area_dwa: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("paved.dwa"))
    )
    paved_area_sewer_pump_capacity: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("paved.tbl"))
    )
    boundaries: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("pluvius.dwa"))
    )
    pluvius: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("pluvius.3b"))
    )
    pluvius_general: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("pluvius.alg"))
    )
    kasklasse: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("kasklass"))
    )
    bui_file: Optional[BuiModel] = Field(None)
    verdampings_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("default.evp"))
    )
    unpaved_area_general: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("unpaved.3b"))
    )
    unpaved_area_storage: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("unpaved.sto"))
    )
    green_house_area_initialisation: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("kasinit"))
    )
    green_house_area_usage_data: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("kasgebr"))
    )
    crop_factors: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("cropfact"))
    )
    table_bergingscoef: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("bergcoef"))
    )
    unpaved_alpha_factor_definitions: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("unpaved.alf"))
    )
    run_log_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("sobek_3b.log"))
    )
    schema_overview: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("3b_gener.out"))
    )
    output_results_paved_area: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("paved.out"))
    )
    output_results_unpaved_area: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("unpaved.out"))
    )
    output_results_green_houses: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("grnhous.out"))
    )
    output_results_open_water: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("openwate.out"))
    )
    output_results_structures: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("struct3b.out"))
    )
    output_results_boundaries: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("bound3b.out"))
    )
    output_results_pluvius: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("pluvius.out"))
    )
    infiltration_definitions_unpaved: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("unpaved.inf"))
    )
    run_debug_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("sobek_3b.dbg"))
    )
    unpaved_seepage: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("unpaved.sep"))
    )
    unpaved_initial_values_tables: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("unpaved.tbl"))
    )
    green_house_general: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("greenhse.3b"))
    )
    green_house_roof_storage: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("greenhse.rf"))
    )
    pluvius_sewage_entry: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("runoff.out"))
    )
    input_variable_gauges_on_edge_nodes: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("sbk_rtc.his"))
    )
    input_salt_data: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("salt.3b"))
    )
    input_crop_factors_open_water: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("crop_ow.prn"))
    )
    restart_file_input: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("RSRR_IN"))
    )
    restart_file_output: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("RSRR_OUT"))
    )
    binary_file_input: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("3b_input.bin"))
    )
    sacramento_input: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("sacrmnto.3b"))
    )
    output_flow_rates_edge_nodes: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("aanvoer.abr"))
    )
    output_salt_concentration_edge: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("saltbnd.out"))
    )
    output_salt_exportation: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("salt.out"))
    )
    greenhouse_silo_definitions: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("greenhse.sil"))
    )
    open_water_general: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("openwate.3b"))
    )
    open_water_seepage_definitions: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("openwate.sep"))
    )
    open_water_tables_target_levels: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("openwate.tbl"))
    )
    structure_general: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("struct3b.dat"))
    )
    structure_definitions: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("struct3b.def"))
    )
    controller_definitions: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("contr3b.def"))
    )
    structure_tables: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("struct3b.tbl"))
    )
    boundary_data: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("bound3b.3b"))
    )
    boundary_tables: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("bound3b.tbl"))
    )
    sobek_location_rtc: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("sbk_loc.rtc"))
    )
    wwtp_data: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("wwtp.3b"))
    )
    wwtp_tables: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("wwtp.tbl"))
    )
    industry_general: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("industry.3b"))
    )

    # Mappix paths should not change and always hold the defined values.
    _validate_mappix_paved_area_sewage_storage = _validator_mappix_value(
        "mappix_paved_area_sewage_storage", _mappix_paved_area_sewage_storage_name
    )
    mappix_paved_area_sewage_storage: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_paved_area_sewage_storage_name)),
        allow_mutation=False,
    )

    _validate_mappix_paved_area_flow_rates = _validator_mappix_value(
        "mappix_paved_area_flow_rates", _mappix_paved_area_flow_rates_name
    )
    mappix_paved_area_flow_rates: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_paved_area_flow_rates_name)),
        allow_mutation=False,
    )
    _validate_mappix_unpaved_area_flow_rates = _validator_mappix_value(
        "mappix_unpaved_area_flow_rates", _mappix_unpaved_area_flow_rates_name
    )
    mappix_unpaved_area_flow_rates: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_unpaved_area_flow_rates_name)),
        allow_mutation=False,
    )
    _validate_mappix_ground_water_levels = _validator_mappix_value(
        "mappix_ground_water_levels", _mappix_ground_water_levels_name
    )
    mappix_ground_water_levels: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_ground_water_levels_name)),
        allow_mutation=False,
    )
    _validate_mappix_green_house_bassins_storage = _validator_mappix_value(
        "mappix_green_house_bassins_storage", _mappix_green_house_bassins_storage_name
    )
    mappix_green_house_bassins_storage: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_green_house_bassins_storage_name)),
        allow_mutation=False,
    )
    _validate_mappix_green_house_bassins_results = _validator_mappix_value(
        "mappix_green_house_bassins_results", _mappix_green_house_bassins_results_name
    )
    mappix_green_house_bassins_results: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_green_house_bassins_results_name)),
        allow_mutation=False,
    )
    _validate_mappix_open_water_details = _validator_mappix_value(
        "mappix_open_water_details", _mappix_open_water_details_name
    )
    mappix_open_water_details: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_open_water_details_name)),
        allow_mutation=False,
    )
    _validate_mappix_exceedance_time_reference_levels = _validator_mappix_value(
        "mappix_exceedance_time_reference_levels",
        _mappix_exceedance_time_reference_levels_name,
    )
    mappix_exceedance_time_reference_levels: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_exceedance_time_reference_levels_name)),
        allow_mutation=False,
    )
    _validate_mappix_flow_rates_over_structures = _validator_mappix_value(
        "mappix_flow_rates_over_structures", _mappix_flow_rates_over_structures_name
    )
    mappix_flow_rates_over_structures: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_flow_rates_over_structures_name)),
        allow_mutation=False,
    )
    _validate_mappix_flow_rates_to_edge = _validator_mappix_value(
        "mappix_flow_rates_to_edge", _mappix_flow_rates_to_edge_name
    )
    mappix_flow_rates_to_edge: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_flow_rates_to_edge_name)),
        allow_mutation=False,
    )
    _validate_mappix_pluvius_max_sewage_storage = _validator_mappix_value(
        "mappix_pluvius_max_sewage_storage", _mappix_pluvius_max_sewage_storage_name
    )
    mappix_pluvius_max_sewage_storage: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_pluvius_max_sewage_storage_name)),
        allow_mutation=False,
    )
    _validate_mappix_pluvius_max_flow_rates = _validator_mappix_value(
        "mappix_pluvius_max_flow_rates", _mappix_pluvius_max_flow_rates_name
    )
    mappix_pluvius_max_flow_rates: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_pluvius_max_flow_rates_name)),
        allow_mutation=False,
    )
    _validate_mappix_balance = _validator_mappix_value(
        "mappix_balance", _mappix_balance_name
    )
    mappix_balance: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_balance_name)),
        allow_mutation=False,
    )
    _validate_mappix_cumulative_balance = _validator_mappix_value(
        "mappix_cumulative_balance", _mappix_cumulative_balance_name
    )
    mappix_cumulative_balance: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_cumulative_balance_name)),
        allow_mutation=False,
    )
    _validate_mappix_salt_concentrations = _validator_mappix_value(
        "mappix_salt_concentrations", _mappix_salt_concentrations_name
    )
    mappix_salt_concentrations: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_salt_concentrations_name)),
        allow_mutation=False,
    )

    industry_tables: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("industry.tbl"))
    )
    maalstop: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("rtc_3b.his"))
    )
    time_series_temperature: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("default.tmp"))
    )
    time_series_runoff: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("rnff.#"))
    )
    discharges_totals_at_edge_nodes: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("bndfltot.his"))
    )
    language_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("sobek_3b.lng"))
    )
    ow_volume: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("ow_vol.his"))
    )
    ow_levels: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("ow_level.his"))
    )
    balance_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("3b_bal.out"))
    )
    his_3B_area_length: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("3bareas.his"))
    )
    his_3B_structure_data: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("3bstrdim.his"))
    )
    his_rr_runoff: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("rrrunoff.his"))
    )
    his_sacramento: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("sacrmnto.his"))
    )
    his_rwzi: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("wwtpdt.his"))
    )
    his_industry: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("industdt.his"))
    )
    ctrl_ini: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("ctrl.ini"))
    )
    root_capsim_input_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("root_sim.inp"))
    )
    unsa_capsim_input_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("unsa_sim.inp"))
    )
    capsim_message_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("capsim.msg"))
    )
    capsim_debug_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("capsim.dbg"))
    )
    restart_1_hour: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("restart1.out"))
    )
    restart_12_hours: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("restart12.out"))
    )
    rr_ready: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("RR-ready"))
    )
    nwrw_areas: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("NwrwArea.His"))
    )
    link_flows: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("3blinks.his"))
    )
    modflow_rr: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("modflow_rr.His"))
    )
    rr_modflow: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("rr_modflow.His"))
    )
    rr_wlm_balance: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("rr_wlmbal.His"))
    )
    sacramento_ascii_output: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("sacrmnto.out"))
    )
    nwrw_input_dwa_table: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("pluvius.tbl"))
    )
    rr_balance: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("rrbalans.his"))
    )
    green_house_classes: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("KasKlasData.dat"))
    )
    green_house_init: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("KasInitData.dat"))
    )
    green_house_usage: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("KasGebrData.dat"))
    )
    crop_factor: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("CropData.dat"))
    )
    crop_ow: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("CropOWData.dat"))
    )
    soil_data: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("SoilData.dat"))
    )
    dio_config_ini_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("dioconfig.ini"))
    )
    bui_file_for_continuous_calculation_series: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("NWRWCONT.#"))
    )
    nwrw_output: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("NwrwSys.His"))
    )
    rr_routing_link_definitions: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("3b_rout.3b"))
    )
    cell_input_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("3b_cel.3b"))
    )
    cell_output_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("3b_cel.his"))
    )
    rr_simulate_log_file: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("sobek3b_progress.txt"))
    )
    rtc_coupling_wq_salt: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("wqrtc.his"))
    )
    rr_boundary_conditions_sobek3: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(Path("BoundaryConditions.bc"))
    )

    rr_ascii_restart_openda: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(filepath=None)
    )
    lgsi_cachefile: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(filepath=None)
    )
    meteo_input_file_rainfall: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(filepath=None)
    )
    meteo_input_file_evaporation: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(filepath=None)
    )
    meteo_input_file_temperature: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(filepath=None)
    )

    @classmethod
    def property_keys(cls) -> Iterable[str]:
        # Skip first two elements corresponding with file_path and serializer_config introduced by the FileModel.
        return list(cls.schema()["properties"].keys())[2:]

    @classmethod
    def _ext(cls) -> str:
        return ".fnm"

    @classmethod
    def _filename(cls) -> str:
        return "rr"

    @classmethod
    def _get_parser(cls) -> Callable[[FilePath], Dict]:
        return lambda path: read(cls.property_keys(), path)

    @classmethod
    def _get_serializer(cls) -> Callable[[Path, Dict, SerializerConfig], None]:
        return write



"""parser.py defines all classes and functions related to parsing pol/pli(z) files.
"""

import warnings
from enum import IntEnum
from pathlib import Path
from typing import Callable, Dict, Iterator, List, Optional, Sequence, Tuple, Union

from hydrolib.core.basemodel import BaseModel
from hydrolib.core.dflowfm.polyfile.models import (
    Description,
    Metadata,
    Point,
    PolyObject,
)


class ParseMsg(BaseModel):
    """ParseMsg defines a single message indicating a significant parse event.

    Attributes:
        line_start (int): The start line of the block to which this ParseMsg refers.
        line_end (int): The end line of the block to which this ParseMsg refers.
        column (Optional[Tuple[int, int]]):
            An optional begin and end column to which this ParseMsg refers.
        reason (str): A human-readable string detailing the reason of the ParseMsg.
    """

    line_start: int
    line_end: int

    column: Optional[Tuple[int, int]]
    reason: str

    def notify_as_warning(self, file_path: Optional[Path] = None):
        """Call warnings.warn with a formatted string describing this ParseMsg

        Args:
            file_path (Optional[Path], optional):
                The file path mentioned in the warning if specified. Defaults to None.
        """
        if self.line_start != self.line_end:
            block_suffix = f"\nInvalid block {self.line_start}:{self.line_end}"
        else:
            block_suffix = f"\nInvalid line {self.line_start}"

        col_suffix = (
            f"\nColumns {self.column[0]}:{self.column[1]}"
            if self.column is not None
            else ""
        )
        file_suffix = f"\nFile: {file_path}" if file_path is not None else ""

        warnings.warn(f"{self.reason}{block_suffix}{col_suffix}{file_suffix}")


class Block(BaseModel):
    """Block is a temporary object which will be converted into a PolyObject.

    The fields are supposed to be set during the lifetime of this object.
    When all fields are set, finalize can be called.

    Attributes:
        start_line (int): The starting line of this current block.
        name (Optional[str]): The name of this block. Defaults to None.
        dimensions (Optional[Tuple[int, int]]):
            The dimensions (n_rows, n_columns) of this Block. Defaults to None.
        points (Optional[List[Point]]):
            The points of this block. Defaults to None.
        ws_warnings (List[ParseMsg]):
            The whitespace warnings associated with this block.
            Defaults to an empty list.
        empty_lines (List[int]):
            The line numbers of the empty lines. Defaults to an empty list.
    """

    start_line: int

    description: Optional[List[str]] = None
    name: Optional[str] = None
    dimensions: Optional[Tuple[int, int]] = None
    points: Optional[List[Point]] = None

    ws_warnings: List[ParseMsg] = []
    empty_lines: List[int] = []

    def finalize(self) -> Optional[Tuple[PolyObject, List[ParseMsg]]]:
        """Finalise this Block and return the constructed PolyObject and warnings

        If the metadata or the points are None, then None is returned.

        Returns:
            Optional[Tuple[PolyObject, List[ParseMsg]]]:
                The constructed PolyObject and warnings encountered while parsing it.
        """
        metadata = self._get_metadata()

        if metadata is None or self.points is None:
            return None

        obj = PolyObject(
            description=self._get_description(), metadata=metadata, points=self.points
        )

        return obj, self.ws_warnings + self._get_empty_line_warnings()

    def _get_description(self) -> Optional[Description]:
        if self.description is not None:
            return Description(content="\n".join(self.description))
        else:
            return None

    def _get_metadata(self) -> Optional[Metadata]:
        if self.name is None or self.dimensions is None:
            return None

        (n_rows, n_columns) = self.dimensions
        return Metadata(name=self.name, n_rows=n_rows, n_columns=n_columns)

    def _get_empty_line_warnings(self):
        if len(self.empty_lines) == 0:
            return []

        warnings = []
        empty_line = (self.empty_lines[0], self.empty_lines[0])

        for line in self.empty_lines[1:]:
            if line == empty_line[1] + 1:
                empty_line = (empty_line[0], line)
            else:
                warnings.append(Block._get_empty_line_msg(empty_line))
                empty_line = (line, line)
        warnings.append(Block._get_empty_line_msg(empty_line))

        return warnings

    @staticmethod
    def _get_empty_line_msg(line_range: Tuple[int, int]) -> ParseMsg:
        return ParseMsg(
            line_start=line_range[0],
            line_end=line_range[1],
            reason="Empty lines are ignored.",
        )


class InvalidBlock(BaseModel):
    """InvalidBlock is a temporary object which will be converted into a ParseMsg.

    Attributes:
        start_line (int): The start line of this InvalidBlock
        end_line (Optional[int]):
            The end line of this InvalidBlock if it is set. Defaults to None.
        invalid_line (int): The line which is causing this block to be invalid.
        reason (str): A human-readable string detailing the reason of the ParseMsg.
    """

    start_line: int
    end_line: Optional[int] = None
    invalid_line: int
    reason: str

    def to_msg(self) -> ParseMsg:
        """Convert this InvalidBlock to the corresponding ParseMsg

        Returns:
            ParseMsg: The ParseMsg corresponding with this InvalidBlock
        """
        return ParseMsg(
            line_start=self.start_line,
            line_end=self.end_line,
            reason=f"{self.reason} at line {self.invalid_line}.",
        )


class ErrorBuilder:
    """ErrorBuilder provides the functionality to the Parser to keep track of errors."""

    def __init__(self) -> None:
        """Create a new ErorrorBuilder"""
        self._current_block: Optional[InvalidBlock] = None

    def start_invalid_block(
        self, block_start: int, invalid_line: int, reason: str
    ) -> None:
        """Start a new invalid block if none exists at the moment

        If we are already in an invalid block, or the previous one
        was never finalised, we will not log the reason, and assume
        it is one long invalid block.

        Args:
            block_start (int): The start of the invalid block.
            invalid_line (int): The actual offending line number.
            reason (str): The reason why this block is invalid.
        """
        if self._current_block is None:
            self._current_block = InvalidBlock(
                start_line=block_start, invalid_line=invalid_line, reason=reason
            )

    def end_invalid_block(self, line: int) -> None:
        """Store the end line of the current block

        If no invalid block currently exists, nothing will be done.

        Args:
            line (int): the final line of this invalid block
        """
        if self._current_block is not None:
            self._current_block.end_line = line

    def finalize_previous_error(self) -> Optional[ParseMsg]:
        """Finalize the current invalid block if it exists

        If no current invalid block exists, None will be returned, and nothing will
        change. If a current block exists, it will be converted into a ParseMsg and
        returned. The current invalid block will be reset.

        Returns:
            Optional[ParseMsg]: The corresponding ParseMsg if an InvalidBlock exists.
        """
        if self._current_block is not None:
            msg = self._current_block.to_msg()
            self._current_block = None

            return msg
        else:
            return None


class StateType(IntEnum):
    """The types of state of a Parser."""

    NEW_BLOCK = 0
    PARSED_DESCRIPTION = 1
    PARSED_NAME = 2
    PARSING_POINTS = 3
    INVALID_STATE = 4


class Parser:
    """Parser provides the functionality to parse a polyfile line by line.

    The Parser parses blocks describing PolyObject instances by relying on
    a rudimentary state machine. The states are encoded with the StateType.
    New lines are fed through the feed_line method. After each line the
    internal state will be updated. When a complete block is read, it will
    be converted into a PolyObject and stored internally.
    When finalise is called, the constructed objects, as well as any warnings
    and errors describing invalid blocks, will be returned.

    Each state defines a feed_line method, stored in the _feed_line dict,
    which consumes a line and potentially transitions the state into the next.
    Each state further defines a finalise method, stored in the _finalise dict,
    which is called upon finalising the parser.

    Invalid states are encoded with INVALID_STATE. In this state the Parser
    attempts to find a new block, and thus looks for a new description or
    name.

    Unexpected whitespace before comments, names, and dimensions, as well as
    empty lines will generate a warning, and will be ignored by the parser.
    """

    def __init__(self, file_path: Path, has_z_value: bool = False) -> None:
        """Create a new Parser

        Args:
            file_path (Path):
                Name of the file being parsed, only used for providing proper warnings.
            has_z_value (bool, optional):
                Whether to interpret the third column as z-coordinates.
                Defaults to False.
        """
        self._has_z_value = has_z_value
        self._file_path = file_path

        self._line = 0
        self._new_block()

        self._error_builder = ErrorBuilder()

        self._poly_objects: List[PolyObject] = []

        self._current_point: int = 0

        self._feed_line: Dict[StateType, Callable[[str], None]] = {
            StateType.NEW_BLOCK: self._parse_name_or_new_description,
            StateType.PARSED_DESCRIPTION: self._parse_name_or_next_description,
            StateType.PARSED_NAME: self._parse_dimensions,
            StateType.PARSING_POINTS: self._parse_next_point,
            StateType.INVALID_STATE: self._parse_name_or_new_description,
        }

        self._finalise: Dict[StateType, Callable[[], None]] = {
            StateType.NEW_BLOCK: self._noop,
            StateType.PARSED_DESCRIPTION: self._add_current_block_as_incomplete_error,
            StateType.PARSED_NAME: self._add_current_block_as_incomplete_error,
            StateType.PARSING_POINTS: self._add_current_block_as_incomplete_error,
            StateType.INVALID_STATE: self._noop,
        }

    def feed_line(self, line: str) -> None:
        """Parse the next line with this Parser.

        Args:
            line (str): The line to parse
        """

        if not Parser._is_empty_line(line):
            self._feed_line[self._state](line)
        else:
            self._handle_empty_line()

        self._increment_line()

    def finalize(self) -> Sequence[PolyObject]:
        """Finalize parsing and return the constructed PolyObject.

        Returns:
            PolyObject:
                A PolyObject containing the constructed PolyObject instances.
        """
        self._error_builder.end_invalid_block(self._line)
        last_error_msg = self._error_builder.finalize_previous_error()
        if last_error_msg is not None:
            self._handle_parse_msg(last_error_msg)

        self._finalise[self._state]()

        return self._poly_objects

    def _new_block(self, offset: int = 0) -> None:
        self._state = StateType.NEW_BLOCK
        self._current_block = Block(start_line=(self._line + offset))

    def _finish_block(self):
        (obj, warnings) = self._current_block.finalize()  # type: ignore
        self._poly_objects.append(obj)

        for msg in warnings:
            self._handle_parse_msg(msg)

        last_error = self._error_builder.finalize_previous_error()
        if last_error is not None:
            self._handle_parse_msg(last_error)

    def _increment_line(self) -> None:
        self._line += 1

    def _noop(self, *_, **__) -> None:
        # no operation
        pass

    def _add_current_block_as_incomplete_error(self) -> None:
        msg = ParseMsg(
            line_start=self._current_block.start_line,
            line_end=self._line,
            reason="EoF encountered before the block is finished.",
        )
        self._handle_parse_msg(msg)

    def _parse_name_or_new_description(self, line: str) -> None:
        if Parser._is_comment(line):
            self._handle_new_description(line)
        elif Parser._is_name(line):
            self._handle_parse_name(line)
        elif self._state != StateType.INVALID_STATE:
            self._handle_new_error("Expected a valid name or description")
            return

        # If we come from an invalid state, and we started a correct new block
        # we will end the previous invalid block, if it exists.
        self._error_builder.end_invalid_block(self._line)

    def _parse_name_or_next_description(self, line: str) -> None:
        if Parser._is_comment(line):
            self._handle_next_description(line)
        elif Parser._is_name(line):
            self._handle_parse_name(line)
        else:
            self._handle_new_error("Expected a valid name or description")

    def _parse_dimensions(self, line: str) -> None:
        dimensions = Parser._convert_to_dimensions(line)

        if dimensions is not None:
            self._current_block.dimensions = dimensions
            self._current_block.points = []
            self._current_point = 0
            self._state = StateType.PARSING_POINTS
        else:
            self._handle_new_error("Expected valid dimensions")

    def _parse_next_point(self, line: str) -> None:
        point = Parser._convert_to_point(
            line, self._current_block.dimensions[1], self._has_z_value  # type: ignore
        )

        if point is not None:
            self._current_block.points.append(point)  # type: ignore
            self._current_point += 1

            if self._current_block.dimensions[0] == self._current_point:  # type: ignore
                self._finish_block()
                self._new_block(offset=1)

        else:
            self._handle_new_error("Expected a valid next point")
            # we parse the line again, as it might be the first line of a new valid
            # block. For example when the invalid block was missing points.
            self._feed_line[self._state](line)

    def _handle_parse_name(self, line: str) -> None:
        self._current_block.name = Parser._convert_to_name(line)
        self._state = StateType.PARSED_NAME

    def _handle_new_description(self, line: str) -> None:
        comment = Parser._convert_to_comment(line)
        self._current_block.description = [
            comment,
        ]
        self._state = StateType.PARSED_DESCRIPTION

    def _handle_next_description(self, line: str) -> None:
        comment = Parser._convert_to_comment(line)
        self._current_block.description.append(comment)  # type: ignore

    def _handle_empty_line(self) -> None:
        if self._state != StateType.INVALID_STATE:
            self._current_block.empty_lines.append(self._line)

    def _handle_new_error(self, reason: str) -> None:
        self._error_builder.start_invalid_block(
            self._current_block.start_line, self._line, reason
        )
        self._state = StateType.INVALID_STATE

    def _handle_parse_msg(self, msg: ParseMsg) -> None:
        msg.notify_as_warning(self._file_path)

    @staticmethod
    def _is_empty_line(line: str) -> bool:
        return len(line.strip()) == 0

    @staticmethod
    def _is_name(line: str) -> bool:
        stripped = line.strip()
        return len(stripped) >= 1 and line[0] != "*" and " " not in stripped

    @staticmethod
    def _convert_to_name(line: str) -> str:
        return line.strip()

    @staticmethod
    def _is_comment(line: str) -> bool:
        return line.strip().startswith("*")

    @staticmethod
    def _convert_to_comment(line: str) -> str:
        return line.strip()[1:]

    @staticmethod
    def _convert_to_dimensions(line: str) -> Optional[Tuple[int, int]]:
        stripped = line.strip()
        elems = stripped.split()

        if len(elems) != 2:
            return None

        try:
            n_rows = int(elems[0])
            n_cols = int(elems[1])

            if n_rows <= 0 or n_cols <= 0:
                return None

            return (n_rows, n_cols)
        except ValueError:
            return None

    @staticmethod
    def _convert_to_point(
        line: str, expected_n_points: int, has_z: bool
    ) -> Optional[Point]:
        stripped = line.strip()
        elems = stripped.split()

        if len(elems) != expected_n_points:
            return None

        try:
            values = list(float(x) for x in elems)

            if has_z:
                x, y, z, *data = values
            else:
                x, y, *data = values
                z = None  # type: ignore

            return Point(x=x, y=y, z=z, data=data)

        except ValueError:
            return None


def _determine_has_z_value(input_val: Union[Path, Iterator[str]]) -> bool:
    return isinstance(input_val, Path) and input_val.suffix == ".pliz"


def read_polyfile(filepath: Path, has_z_values: Optional[bool] = None) -> Dict:
    """Read the specified file and return the corresponding data.

    The file is expected to follow the .pli(z) / .pol convention. A .pli(z) or .pol
    file is defined as consisting of a number of blocks of lines adhering to the
    following format:

    - Optional description record consisting of one or more lines starting with '*'.
        These will be ignored.
    - Name consisting of a non-blank character string
    - Two integers, Nr and Nc, representing the numbers of rows and columns respectively
    - Nr number of data points, consisting of Nc floats separated by whitespace

    For example:
    ```
    ...
    *
    * Polyline L008
    *
    L008
    4 2
        131595.0 549685.0
        131750.0 549865.0
        131595.0 550025.0
        131415.0 550175.0
    ...
    ```

    Note that the points can be arbitrarily indented, and the comments are optional.

    if no has_z_value has been defined, it will be based on the file path
    extensions of the filepath:
    - .pliz will default to True
    - .pli and .pol will default to False

    Empty lines and unexpected whitespace will be flagged as warnings, and ignored.

    If invalid syntax is detected within a block, an error will be created. This block
    will be ignored for the purpose of creating PolyObject instances.
    Once an error is encountered, any following lines will be marked as part of the
    invalid block, until a new valid block is found. Note that this means that sequential
    invalid blocks will be reported as a single invalid block. Such invalid blocks will
    be reported as warnings.

    Args:
        filepath:
            Path to the pli(z)/pol convention structured file.
        has_z_values:
            Whether to create points containing a z-value. Defaults to None.

    Returns:
        Dict: The dictionary describing the data of a PolyObject.
    """
    if has_z_values is None:
        has_z_values = _determine_has_z_value(filepath)

    parser = Parser(filepath, has_z_value=has_z_values)

    with filepath.open("r") as f:
        for line in f:
            parser.feed_line(line)

    objs = parser.finalize()

    return {"has_z_values": has_z_values, "objects": objs}



import logging
from enum import Enum
from pathlib import Path
from typing import List, Literal, Optional

from pydantic import Field, NonNegativeInt, PositiveInt
from pydantic.class_validators import validator

from hydrolib.core.basemodel import (
    DiskOnlyFileModel,
    validator_set_default_disk_only_file_model_when_none,
)
from hydrolib.core.dflowfm.ini.models import INIBasedModel, INIGeneral, INIModel
from hydrolib.core.dflowfm.ini.util import (
    get_enum_validator,
    get_split_string_on_delimiter_validator,
    make_list_validator,
)

logger = logging.getLogger(__name__)


class FrictionType(str, Enum):
    """
    Enum class containing the valid values for the frictionType
    attribute in several subclasses of Structure/CrossSection/friction.models.
    """

    chezy = "Chezy"
    """str: Chézy C [m 1/2 /s]"""

    manning = "Manning"
    """str: Manning n [s/m 1/3 ]"""

    walllawnikuradse = "wallLawNikuradse"
    """str: Nikuradse k_n [m]"""

    whitecolebrook = "WhiteColebrook"
    """str: Nikuradse k_n [m]"""

    stricklernikuradse = "StricklerNikuradse"
    """str: Nikuradse k_n [m]"""

    strickler = "Strickler"
    """str: Strickler k_s [m 1/3 /s]"""

    debosbijkerk = "deBosBijkerk"
    """str: De Bos-Bijkerk γ [1/s]"""


class FrictGeneral(INIGeneral):
    """The friction file's `[General]` section with file meta data."""

    class Comments(INIBasedModel.Comments):
        fileversion: Optional[str] = Field(
            "File version. Do not edit this.", alias="fileVersion"
        )
        filetype: Optional[str] = Field(
            "File type. Should be 'roughness'. Do not edit this.",
            alias="fileType",
        )
        frictionvaluesfile: Optional[str] = Field(
            "Name of <*.bc> file containing the timeseries with friction values. "
            + "Only needed for functionType = timeSeries.",
            alias="frictionValuesFile",
        )

    comments: Comments = Comments()
    _header: Literal["General"] = "General"
    fileversion: str = Field("3.01", alias="fileVersion")
    filetype: Literal["roughness"] = Field("roughness", alias="fileType")
    frictionvaluesfile: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(None), alias="frictionValuesFile"
    )

    _disk_only_file_model_should_not_be_none = (
        validator_set_default_disk_only_file_model_when_none()
    )


class FrictGlobal(INIBasedModel):
    """A `[Global]` block for use inside a friction file.

    Multiple of such blocks may be present to define multiple frictionId classes.
    """

    class Comments(INIBasedModel.Comments):
        frictionid: Optional[str] = Field(
            "Name of the roughness variable.", alias="frictionId"
        )
        frictiontype: Optional[str] = Field(
            "The global roughness type for this variable, which is used "
            + "if no branch specific roughness definition is given.",
            alias="frictionType",
        )
        frictionvalue: Optional[str] = Field(
            "The global default value for this roughness variable.",
            alias="frictionValue",
        )

    comments: Comments = Comments()
    _header: Literal["Global"] = "Global"
    frictionid: str = Field(alias="frictionId")
    frictiontype: FrictionType = Field(alias="frictionType")
    frictionvalue: float = Field(alias="frictionValue")

    _frictiontype_validator = get_enum_validator("frictiontype", enum=FrictionType)

    def _get_identifier(self, data: dict) -> Optional[str]:
        return data.get("frictionid")


class FrictBranch(INIBasedModel):
    """A `[Branch]` block for use inside a friction file.

    Each block can define the roughness value(s) on a particular branch.
    """

    class Comments(INIBasedModel.Comments):
        branchid: Optional[str] = Field("The name of the branch.", alias="branchId")
        frictiontype: Optional[str] = Field(
            "The roughness type to be used on this branch.", alias="frictionType"
        )
        functiontype: Optional[str] = Field(
            "Function type for the calculation of the value. "
            + "Possible values: constant, timeSeries, absDischarge, waterlevel.",
            alias="functionType",
        )
        timeseriesid: Optional[str] = Field(
            "Refers to a data block in the <*.bc> frictionValuesFile. "
            + "Only if functionType = timeSeries.",
            alias="timeSeriesId",
        )
        numlevels: Optional[str] = Field(
            "Number of levels in table. Only if functionType is not constant.",
            alias="numLevels",
        )
        levels: Optional[str] = Field(
            "Space separated list of discharge [m3/s] or water level [m AD] values. "
            + "Only if functionType is absDischarge or waterLevel."
        )
        numlocations: Optional[str] = Field(
            "Number of locations on branch. The default 0 implies branch uniform values.",
            alias="numLocations",
        )
        chainage: Optional[str] = Field(
            "Space separated list of locations on the branch [m]. Locations sorted by "
            + "increasing chainage. The keyword must be specified if numLocations>0."
        )
        frictionvalues: Optional[str] = Field(
            "numLevels lines containing space separated lists of roughness values: "
            + "numLocations values per line. If the functionType is constant, then a "
            + "single line is required. For a uniform roughness per branch "
            + "(numLocations = 0) a single entry per line is required. The meaning "
            + "of the values depends on the roughness type selected (see frictionType).",
            alias="frictionValues",
        )

    comments: Comments = Comments()
    _header: Literal["Branch"] = "Branch"
    branchid: str = Field(alias="branchId")
    frictiontype: FrictionType = Field(alias="frictionType")
    functiontype: Optional[str] = Field("constant", alias="functionType")
    timeseriesid: Optional[str] = Field(alias="timeSeriesId")
    numlevels: Optional[PositiveInt] = Field(alias="numLevels")
    levels: Optional[List[float]]
    numlocations: Optional[NonNegativeInt] = Field(0, alias="numLocations")
    chainage: Optional[List[float]]
    frictionvalues: Optional[List[float]] = Field(
        alias="frictionValues"
    )  # TODO: turn this into List[List[float]], see issue #143.

    _split_to_list = get_split_string_on_delimiter_validator(
        "levels",
        "chainage",
        "frictionvalues",
    )

    _frictiontype_validator = get_enum_validator("frictiontype", enum=FrictionType)

    def _get_identifier(self, data: dict) -> Optional[str]:
        return data.get("branchid")

    @validator("levels", always=True)
    @classmethod
    def _validate_levels(cls, v, values):
        if v is not None and (
            values["numlevels"] is None or len(v) != values["numlevels"]
        ):
            raise ValueError(
                f"Number of values for levels should be equal to the numLevels value (branchId={values.get('branchid', '')})."
            )

        return v

    @validator("chainage", always=True)
    @classmethod
    def _validate_chainage(cls, v, values):
        if v is not None and len(v) != values["numlocations"]:
            raise ValueError(
                f"Number of values for chainage should be equal to the numLocations value (branchId={values.get('branchid', '')})."
            )

        return v

    @validator("frictionvalues", always=True)
    @classmethod
    def _validate_frictionvalues(cls, v, values):
        # number of values should be equal to numlocations*numlevels
        numlevels = (
            1
            if (
                "numlevels" not in values
                or values["numlevels"] is None
                or values["numlevels"] == 0
            )
            else values["numlevels"]
        )
        numvals = max(1, values["numlocations"]) * numlevels
        if v is not None and len(v) != numvals:
            raise ValueError(
                f"Number of values for frictionValues should be equal to the numLocations*numLevels value (branchId={values.get('branchid', '')})."
            )

        return v


class FrictionModel(INIModel):
    """
    The overall friction model that contains the contents of one friction file.

    This model is typically referenced under a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.frictfile[..]`.

    Attributes:
        general (FrictGeneral): `[General]` block with file metadata.
        global_ (List[FrictGlobal]): Definitions of `[Global]` friction classes.
        branch (List[FrictBranch]): Definitions of `[Branch]` friction values.
    """

    general: FrictGeneral = FrictGeneral()
    global_: List[FrictGlobal] = Field([], alias="global")  # to circumvent built-in kw
    branch: List[FrictBranch] = []

    _split_to_list = make_list_validator(
        "global_",
        "branch",
    )

    @classmethod
    def _ext(cls) -> str:
        return ".ini"

    @classmethod
    def _filename(cls) -> str:
        return "roughness"



from .models import (
    DIMR,
    Component,
    ComponentOrCouplerRef,
    ControlModel,
    CoupledItem,
    Coupler,
    Documentation,
    FMComponent,
    GlobalSettings,
    KeyValuePair,
    Logger,
    Parallel,
    RRComponent,
    Start,
    StartGroup,
)

__all__ = [
    "KeyValuePair",
    "Component",
    "FMComponent",
    "RRComponent",
    "Documentation",
    "GlobalSettings",
    "ComponentOrCouplerRef",
    "CoupledItem",
    "Logger",
    "Coupler",
    "StartGroup",
    "ControlModel",
    "Parallel",
    "Start",
    "DIMR",
]



from typing import Dict, List, Literal, Optional

from pydantic.class_validators import root_validator
from pydantic.fields import Field

from hydrolib.core.dflowfm.ini.models import INIBasedModel, INIGeneral, INIModel
from hydrolib.core.dflowfm.ini.util import (
    LocationValidationConfiguration,
    get_split_string_on_delimiter_validator,
    validate_location_specification,
)


class ObservationCrossSectionGeneral(INIGeneral):
    """The observation cross section file's `[General]` section with file meta data."""

    class Comments(INIBasedModel.Comments):
        fileversion: Optional[str] = Field(
            "File version. Do not edit this.", alias="fileVersion"
        )
        filetype: Optional[str] = Field(
            "File type. Should be 'obsCross'. Do not edit this.", alias="fileType"
        )

    comments: Comments = Comments()
    fileversion: str = Field("2.00", alias="fileVersion")
    filetype: Literal["obsCross"] = Field("obsCross", alias="fileType")


class ObservationCrossSection(INIBasedModel):
    """
    The observation cross section that is included in the
    observation cross section file.

    All lowercased attributes match with the observation cross
    section output as described in [UM Sec.F2.4.1]
    (https://content.oss.deltares.nl/delft3d/manuals/D-Flow_FM_User_Manual_1D2D.pdf#subsubsection.F.2.4.1)
    """

    class Comments(INIBasedModel.Comments):
        name: Optional[str] = "Name of the cross section (max. 255 characters)."
        branchid: Optional[str] = Field(
            "(optional) Branch on which the cross section is located.", alias="branchId"
        )
        chainage: Optional[str] = "(optional) Location on the branch (m)."
        numcoordinates: Optional[str] = Field(
            "(optional) Number of values in xCoordinates and yCoordinates. "
            "This value should be greater than or equal to 2.",
            alias="numCoordinates",
        )
        xcoordinates: Optional[str] = Field(
            "(optional) x-coordinates of the cross section line. "
            "(number of values = numCoordinates)",
            alias="xCoordinates",
        )
        ycoordinates: Optional[str] = Field(
            "(optional) y-coordinates of the cross section line. "
            "(number of values = numCoordinates)",
            alias="yCoordinates",
        )

    comments: Comments = Comments()
    _header: Literal["ObservationCrossSection"] = "ObservationCrossSection"
    name: str = Field(max_length=255, alias="name")
    branchid: Optional[str] = Field(alias="branchId")
    chainage: Optional[float] = Field(alias="chainage")
    numcoordinates: Optional[int] = Field(alias="numCoordinates")
    xcoordinates: Optional[List[float]] = Field(alias="xCoordinates")
    ycoordinates: Optional[List[float]] = Field(alias="yCoordinates")

    _split_to_list = get_split_string_on_delimiter_validator(
        "xcoordinates", "ycoordinates"
    )

    @root_validator(allow_reuse=True)
    def validate_that_location_specification_is_correct(cls, values: Dict) -> Dict:
        """Validates that the correct location specification is given."""
        return validate_location_specification(
            values,
            config=LocationValidationConfiguration(
                validate_node=False, minimum_num_coordinates=2
            ),
        )

    def _get_identifier(self, data: dict) -> Optional[str]:
        return data.get("name")


class ObservationCrossSectionModel(INIModel):
    """
    The overall observation cross section model that contains the contents
    of one observation cross section file.
    """

    general: ObservationCrossSectionGeneral = ObservationCrossSectionGeneral()
    observationcrosssection: List[ObservationCrossSection] = []



from pathlib import Path
from typing import Callable, Dict, List, Optional

from pydantic import Field

from hydrolib.core.basemodel import BaseModel, ParsableFileModel, SerializerConfig

from .parser import XYZParser
from .serializer import XYZSerializer


class XYZPoint(BaseModel):
    """Single sample or forcing point.

    Attributes:
        x: x or λ coordinate
        y: y or φ coordinate
        z: sample value or group number (forcing)
        comment: keyword for grouping (forcing)
    """

    x: float
    y: float
    z: float
    comment: Optional[str] = Field(
        None, alias="group", description="comment or group name"
    )

    def _get_identifier(self, data: dict) -> Optional[str]:
        x = data.get("x")
        y = data.get("y")
        z = data.get("z")
        return f"x:{x} y:{y} z:{z}"


class XYZModel(ParsableFileModel):
    """Sample or forcing file.

    Attributes:
        points: List of [`XYZPoint`][hydrolib.core.dflowfm.xyz.models.XYZPoint]
    """

    points: List[XYZPoint]

    def dict(self, *args, **kwargs):
        # speed up serializing by not converting these lowest models to dict
        return dict(points=self.points)

    @classmethod
    def _ext(cls) -> str:
        return ".xyz"

    @classmethod
    def _filename(cls) -> str:
        return "sample"

    @classmethod
    def _get_serializer(cls) -> Callable[[Path, Dict, SerializerConfig], None]:
        return XYZSerializer.serialize

    @classmethod
    def _get_parser(cls) -> Callable:
        return XYZParser.parse



from __future__ import annotations

import json
import logging
from collections import namedtuple
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple

import netCDF4 as nc
import numpy as np

from hydrolib.core.basemodel import BaseModel

if TYPE_CHECKING:
    from ..dflowfm.net.models import Link1d2d, Mesh1d, Mesh2d


class UgridReader:
    """UgridReader provides the logic to read a specified UGRID file."""

    def __init__(self, file_path: Path) -> None:
        """Creates a new UgridReader, reading the specified path.

        Args:
            file_path (Path): The path to read from.

        Raises:
            OSError: Thrown when file_path does not exist.
        """
        self._ncfile_path = file_path

        if not self._ncfile_path.exists():
            raise OSError(f'File "{self._ncfile_path}" not found.')

        self._explorer = NCExplorer.from_file_path(self._ncfile_path)

    def read_mesh1d_network1d(self, mesh1d: Mesh1d) -> None:
        """
        Read the Ugrid from the netcdf and add the dflowfm cstructure with grid to the
        specified mesh1d.

        Args:
            mesh1d (Mesh1d): The object to which the read network1d is added.
        """
        if (
            self._explorer.mesh1d_var_name_mapping is None
            or self._explorer.network1d_var_name_mapping is None
        ):
            logging.debug(
                "Mesh1d and Network are not found in the dataset, reading is skipped."
            )
            return

        # If the mesh is not given (default), use the networks one
        ds = nc.Dataset(self._ncfile_path)  # type: ignore[import]

        # Read mesh1d
        for meshkey, nckey in self._explorer.mesh1d_var_name_mapping.items():
            setattr(mesh1d, meshkey, self._read_nc_attribute(ds[nckey]))

        # Read network variables
        for meshkey, nckey in self._explorer.network1d_var_name_mapping.items():
            setattr(mesh1d, meshkey, self._read_nc_attribute(ds[nckey]))

        # Process network
        mesh1d._process_network1d()

        ds.close()

    def read_mesh2d(self, mesh2d: Mesh2d) -> None:
        """
        Read the Ugrid from the netcdf and add the dflowfm cstructure with grid to the
        specified mesh2d.

        Args:
            mesh2d (Mesh2d): The object to which the read network1d is added.
        """
        if self._explorer.mesh2d_var_name_mapping is None:
            logging.debug("Mesh2d is not found in the dataset, reading is skipped.")
            return

        ds = nc.Dataset(self._ncfile_path)  # type: ignore[import]

        # Read mesh1d
        for meshkey, nckey in self._explorer.mesh2d_var_name_mapping.items():
            setattr(mesh2d, meshkey, self._read_nc_attribute(ds[nckey]))

        ds.close()

    def read_link1d2d(self, link1d2d: Link1d2d) -> None:
        """Read the Link1d2d from the wrapped netCDF file of this UgridReader.

        Args:
            link1d2d (Link1d2d): The Link1d2d to which the data is added.
        """
        if self._explorer.link1d2d_var_name_mapping is None:
            logging.debug("Link1d2d is not found in the dataset, reading is skipped.")
            return

        ds = nc.Dataset(self._ncfile_path)  # type: ignore[import]

        # Read mesh1d
        for meshkey, nckey in self._explorer.link1d2d_var_name_mapping.items():
            setattr(link1d2d, meshkey, self._read_nc_attribute(ds[nckey]))

        ds.close()

    def _read_nc_attribute(self, attr: nc._netCDF4.Variable) -> np.ndarray:
        """Read values from netcdf attribute.
        - Character arrays are converted to strings
        - Masked arrays are converted to regular arrays where the masked values are filled with the fillvalue
        - The numerical arrays are subtracted by the start value. This is often 1, while python is 0-based.

        Args:
            attr (netCDF4._netCDF4.Variable): Attribute in the file

        Returns:
            np.ndarray: returned array
        """
        values = attr[:]
        if values.dtype == "S1":
            # Convert to strings
            arr = np.array(list(map(str.strip, nc.chartostring(values.data))))  # type: ignore[import]

        else:
            # Get data from masked array
            arr = values[:].data
            # Fill masked values with fillvalue
            if values.mask.any():
                arr[values.mask] = attr._FillValue
            # Substract start index
            if hasattr(attr, "start_index"):
                arr -= attr.start_index

        return arr


class NCExplorer(BaseModel):
    """NCExplorer provides the mapping of the UGRID variable names as used within
    HYDROLIB models to the actual values used within the netCDF file.

    If a component is not present, the corresponding mapping is set to None.
    A NCExplorer can be constructed from a file by using the `from_file_path`
    class method.

    Attributes:
        network1d_var_name_mapping (Optional[Dict[str, str]]):
            The mapping of Network variable names.
        mesh1d_var_name_mapping (Optional[Dict[str, str]]):
            The mapping of Mesh1d variable names.
        mesh2d_var_name_mapping (Optional[Dict[str, str]]):
            The mapping of Mesh2d variable names.
        link1d2d_var_name_mapping (Optional[Dict[str, str]]):
            The mapping of Link1d2d variable names.
    """

    Keys = namedtuple("Keys", ["network1d", "mesh1d", "mesh2d", "link1d2d"])

    network1d_var_name_mapping: Optional[Dict[str, str]]
    mesh1d_var_name_mapping: Optional[Dict[str, str]]
    mesh2d_var_name_mapping: Optional[Dict[str, str]]
    link1d2d_var_name_mapping: Optional[Dict[str, str]]

    @classmethod
    def from_file_path(cls, file_path: Path) -> "NCExplorer":
        """Create a new NCExplorer from the specified file_path.

        Args:
            file_path (Path): The path to the net.nc file.

        Returns:
            NCExplorer: A newly initialized NCExplorer.
        """
        conventions = NCExplorer._read_ugrid_conventions()
        dataset = nc.Dataset(file_path)  # type: ignore[import]

        keys = NCExplorer._determine_keys(dataset)
        network1d_mapping = NCExplorer._retrieve_variable_names_mapping(
            keys.network1d, dataset, conventions["network1d"]
        )
        mesh1d_mapping = NCExplorer._retrieve_variable_names_mapping(
            keys.mesh1d, dataset, conventions["mesh1d"]
        )
        mesh2d_mapping = NCExplorer._retrieve_variable_names_mapping(
            keys.mesh2d, dataset, conventions["mesh2d"]
        )
        link1d2d_mapping = NCExplorer._retrieve_variable_names_mapping(
            keys.link1d2d, dataset, conventions["link1d2d"]
        )

        dataset.close()

        return cls(
            network1d_var_name_mapping=network1d_mapping,
            mesh1d_var_name_mapping=mesh1d_mapping,
            mesh2d_var_name_mapping=mesh2d_mapping,
            link1d2d_var_name_mapping=link1d2d_mapping,
        )

    @staticmethod
    def _read_ugrid_conventions() -> Dict:
        with open(Path(__file__).parent.joinpath("ugrid_conventions.json"), "r") as f:
            return json.load(f)

    @staticmethod
    def _determine_keys(dataset: nc.Dataset) -> NCExplorer.Keys:  # type: ignore[import]
        keys1d = NCExplorer._retrieve_1d_keys(dataset)

        mesh1d = keys1d[0] if keys1d is not None else None
        network1d = keys1d[1] if keys1d is not None else None

        mesh2d = NCExplorer._retrieve_2d_key(dataset)
        link1d2d = NCExplorer._retrieve_1d2d_key(dataset)

        return NCExplorer.Keys(network1d, mesh1d, mesh2d, link1d2d)

    @staticmethod
    def _retrieve_1d_keys(dataset) -> Optional[Tuple[str, str]]:
        def is_mesh1d(ncdata) -> bool:
            return (
                NCExplorer._has_cf_role(ncdata, "mesh_topology")
                and NCExplorer._has_n_topology_dimension(ncdata, 1)
                and NCExplorer._has_attribute(ncdata, "coordinate_space")
            )

        for var_key, ncdata in dataset.variables.items():
            if is_mesh1d(ncdata):
                return var_key, ncdata.getncattr("coordinate_space")
        return None

    @staticmethod
    def _retrieve_2d_key(dataset) -> Optional[str]:
        def is_mesh2d(ncdata) -> bool:
            return NCExplorer._has_cf_role(
                ncdata, "mesh_topology"
            ) and NCExplorer._has_n_topology_dimension(ncdata, 2)

        for var_key, ncdata in dataset.variables.items():
            if is_mesh2d(ncdata):
                return var_key
        return None

    @staticmethod
    def _retrieve_1d2d_key(dataset) -> Optional[str]:
        def is_link1d2d(ncdata) -> bool:
            return NCExplorer._has_cf_role(ncdata, "mesh_topology_contact")

        for var_key, ncdata in dataset.variables.items():
            if is_link1d2d(ncdata):
                return var_key
        return None

    @staticmethod
    def _has_cf_role(ncdata, cf_role_value: str) -> bool:
        return NCExplorer._has_attribute(ncdata, "cf_role", cf_role_value)

    @staticmethod
    def _has_n_topology_dimension(ncdata, n: int) -> bool:
        return NCExplorer._has_attribute(ncdata, "topology_dimension", n)

    @staticmethod
    def _has_attribute(ncdata, name: str, value: Any = None):
        attributes = ncdata.ncattrs()
        return name in attributes and (value is None or ncdata.getncattr(name) == value)

    @staticmethod
    def _retrieve_variable_names_mapping(
        variable_key: Optional[str], dataset, conventions: Dict
    ) -> Optional[Dict]:
        if variable_key is None:
            # particular key does not exist, as such there are no variables either.
            return None

        nc_variables_set = set(dataset.variables.keys())

        result = {}

        for key, value in conventions.items():
            nc_vars = set(variable_key + suffix for suffix in value["suffices"])
            in_dataset = nc_vars & nc_variables_set

            len_in_data_set = len(in_dataset)

            if len_in_data_set == 1:
                result[key] = in_dataset.pop()
            elif len_in_data_set > 1:
                raise KeyError(
                    f'Multiple attributes for "{key}" were found in nc file. Got "{" and ".join(in_dataset)}"'
                )
            elif "is_optional" in value and value["is_optional"]:
                logging.info(
                    "Optional variable '%s' is not present in the nc_file and is skipped",
                    key,
                )
            else:
                raise KeyError(
                    f'An attribute for "{key}" was not found in nc file. Expected "{"/".join(nc_vars)}"'
                )

        return result



from .models import (
    T3D,
    Astronomic,
    AstronomicCorrection,
    Constant,
    ForcingGeneral,
    ForcingModel,
    Harmonic,
    HarmonicCorrection,
    QHTable,
    QuantityUnitPair,
    RealTime,
    TimeInterpolation,
    TimeSeries,
    VectorQuantityUnitPairs,
    VerticalInterpolation,
    VerticalPositionType,
)

__all__ = [
    "VerticalInterpolation",
    "VerticalPositionType",
    "TimeInterpolation",
    "QuantityUnitPair",
    "VectorQuantityUnitPairs",
    "TimeSeries",
    "Harmonic",
    "Astronomic",
    "HarmonicCorrection",
    "AstronomicCorrection",
    "T3D",
    "QHTable",
    "Constant",
    "ForcingGeneral",
    "ForcingModel",
    "RealTime",
]



import platform
import re
from enum import Enum, auto
from operator import eq, ge, gt, le, lt, ne
from typing import Any, Callable, List, Optional


def example(a: float, b: float = 1.0) -> float:
    """[summary]

    Args:
        a (float): [description]
        b (float): [description]

    Returns:
        float: [description]
    """
    return a * b


def to_key(string: str) -> str:
    """
    Construct a key name from a given field name.
    The given field name may be a Pydantic Field alias, and the key name
    is intended to be used as a BaseModel class member variable name.

    Args:
        string (str): input field name

    """
    # First replace any leading digits, because those are undesirable
    # in variable names.
    digitdict = {
        "0": "zero",
        "1": "one",
        "2": "two",
        "3": "three",
        "4": "four",
        "5": "five",
        "6": "six",
        "7": "seven",
        "8": "eight",
        "9": "nine",
    }
    m = re.search(r"^\d+", string)
    if m:
        digitstring = string[0 : m.end()]
        for key, val in digitdict.items():
            digitstring = digitstring.replace(key, val)
        string = digitstring + string[m.end() :]

    # Next, replace spaces and hyphens in the potential variable name.
    return string.lower().replace(" ", "_").replace("-", "")


def to_list(item: Any) -> List[Any]:
    """Puts the specified item in a list if it is an instance of `dict`.

    Attributes:
        item: The item to put in a list.

    Returns:
        List: A list with the specified item.
    """

    if not isinstance(item, list):
        return [item]
    return item


def str_is_empty_or_none(str_field: str) -> bool:
    """
    Verifies whether a string is empty or None.

    Args:
        str_field (str): String to validate.

    Returns:
        bool: Evaluation result.
    """
    return str_field is None or not str_field or str_field.isspace()


def get_str_len(str_field: Optional[str]) -> int:
    """
    Get string length or 0 if input is None.

    Args:
        str_field (str): String to measure.

    Returns:
        int: Length of passed input.
    """
    return len(str_field) if str_field else 0


def get_substring_between(source: str, start: str, end: str) -> Optional[str]:
    """Finds the substring between two other strings.

    Args:
        source (str): The source string.
        start (str): The starting string from where to create the substring.
        end (str): The end string to where to create the substring.

    Returns:
        str: The substring if found; otherwise, `None`.
    """

    index_start = source.find(start)
    if index_start == -1:
        return None

    index_start += len(start)

    index_end = source.rfind(end, index_start)
    if index_end == -1:
        return None

    return source[index_start:index_end]


def operator_str(operator_func: Callable) -> str:
    """
    Make string representation of some of operator's built-in operator
    functions, for use in prettyprinting.

    Args:
        operator_func (Callable): Typically one of operator's built-in
            operator functions. When unsupported, the standard __str__
            representation is returned.
    """
    if operator_func == eq:
        return "is"
    elif operator_func == ne:
        return "is not"
    elif operator_func == lt:
        return "is less than"
    elif operator_func == le:
        return "is less than or equal to"
    elif operator_func == gt:
        return "is greater than"
    elif operator_func == ge:
        return "is greater than or equal to"
    else:
        return str(operator_func)


class OperatingSystem(Enum):
    """Enum represting an operating system."""

    WINDOWS = auto()
    """The Windows operating system."""
    LINUX = auto()
    """The Linux operating system."""
    MACOS = auto()
    """The MacOS operating system."""


def get_operating_system() -> OperatingSystem:
    """Gets the currently running operating system.

    Raises:
        NotImplementedError: When the current operating system is anything other than Windows, Linux or MacOS.

    Returns:
        OperatingSystem: The operating system.
    """
    operating_system = platform.system()

    if operating_system == "Windows":
        return OperatingSystem.WINDOWS
    if operating_system == "Linux":
        return OperatingSystem.LINUX
    if operating_system == "Darwin":
        return OperatingSystem.MACOS

    raise NotImplementedError(f"Operating system {operating_system} is not supported.")
